import astropy.units
import matplotlib.pyplot as plt
import numpy as np
import pytest

from mcalf.utils.plot import calculate_axis_extent, calculate_extent, class_cmap, hide_existing_labels

from ..helpers import figure_test


def test_hide_existing_labels():
    plot_settings = {
        'LineA': {'color': 'r', 'label': 'A'},
        'LineB': {'color': 'g', 'label': 'B'},
        'LineC': {'color': 'b', 'label': 'C'},
    }
    fig, ax = plt.subplots()
    ax.plot([0, 1], [0, 1], **plot_settings['LineA'])
    ax.plot([0, 1], [1, 0], **plot_settings['LineB'])
    hide_existing_labels(plot_settings)
    plt.close()
    ret = [x['label'] for x in plot_settings.values()]
    assert ret == ['_A', '_B', 'C']


def test_calculate_axis_extent():
    # Invalid `res`
    with pytest.raises(TypeError) as e:
        calculate_axis_extent('1 m', 1000)
    assert '`resolution` values must be' in str(e.value)

    # Invalid `px`
    with pytest.raises(TypeError) as e:
        calculate_axis_extent(250.0, 1000.5)
    assert '`px` must be an integer' in str(e.value)

    # Invalid `offset`
    with pytest.raises(TypeError) as e:
        calculate_axis_extent(250.0, 1000, [10])
    assert '`offset` must be an float or integer' in str(e.value)

    # Default `unit` used
    _, _, un = calculate_axis_extent(250.0, 1000, 10)
    assert un == 'Mm'  # Will fail if default value is changed
    _, _, un = calculate_axis_extent(250.0, 1000, 10, 'testunit')
    assert un == 'testunit'  # Will fail if default value is changed

    # Unit extracted
    _, _, un = calculate_axis_extent(250.0 * astropy.units.kg, 1000, 10, 'testunit')
    assert un == '$\\mathrm{kg}$'

    # Test good values
    f, l, _ = calculate_axis_extent(3., 7, -3.5)
    assert -10.5 == pytest.approx(f)
    assert 10.5 == pytest.approx(l)
    f, l, _ = calculate_axis_extent(2., 2, 1)
    assert 2. == pytest.approx(f)
    assert 6. == pytest.approx(l)


def test_calculate_extent():

    # Resolution of None should return None
    assert calculate_extent((100, 200), None) is None

    # Shape is 2-tuple
    for a in ((1, 1, 1), np.array([1, 1])):
        with pytest.raises(TypeError) as e:
            calculate_extent(a, resolution=(1, 1), offset=(1, 1))
        assert '`shape` must be a tuple of length 2' in str(e.value)

    # Resolution is 2-tuple
    for a in ((1, 1, 1), np.array([1, 1])):
        with pytest.raises(TypeError) as e:
            calculate_extent((100, 200), resolution=a, offset=(1, 1))
        assert '`resolution` must be a tuple of length 2' in str(e.value)

    # Offset is 2-tuple
    for a in ((1, 1, 1), np.array([1, 1])):
        with pytest.raises(TypeError) as e:
            calculate_extent((100, 200), resolution=(1, 1), offset=a)
        assert '`offset` must be a tuple of length 2' in str(e.value)

    # Test good value
    l, r, b, t = calculate_extent((7, 2), (2., 3.), (1, -3.5))
    assert 2. == pytest.approx(l)
    assert 6. == pytest.approx(r)
    assert -10.5 == pytest.approx(b)
    assert 10.5 == pytest.approx(t)


def plot_helper_calculate_extent(*args, dimension=None):
    fig, ax = plt.subplots()
    calculate_extent(*args, ax=ax, dimension=dimension)
    x, y = ax.get_xlabel(), ax.get_ylabel()
    plt.close(fig)
    return x, y


def test_calculate_extent_ax():

    # Common args
    args = ((7, 2), (2., 3.), (1, -3.5))

    # Dimension is 2-tuple or 2-list
    for d in ((1, 1, 1), np.array([1, 1]), [1, 1, 1]):
        with pytest.raises(TypeError) as e:
            plot_helper_calculate_extent(*args, dimension=d)
        assert '`dimension` must be a tuple or list of length 2' in str(e.value)

    # Different for both
    x, y = plot_helper_calculate_extent(*args, dimension=('test / one', 'test / two'))
    assert 'test / one (Mm)' == x
    assert 'test / two (Mm)' == y

    # Default values
    x, y = plot_helper_calculate_extent(*args)
    assert 'x-axis (Mm)' == x
    assert 'y-axis (Mm)' == y

    # Same for both
    x, y = plot_helper_calculate_extent(*args, dimension='test / same')
    assert 'test / same (Mm)' == x
    assert 'test / same (Mm)' == y


def test_class_cmap():

    # Incorrect `n` type
    with pytest.raises(TypeError) as e:
        class_cmap('original', 2.5)
    assert '`n` must be an integer' in str(e.value)

    # Correct `n` type
    class_cmap('original', 4)


@figure_test
def test_class_cmap_plot(pytestconfig):

    params = (
        ('original', 3),
        ('original', 5),
        ('original', 6),
        ('jet', 4),
    )

    fig, axes = plt.subplots(len(params))

    for ax, (style, n) in zip(axes, params):
        gradient = np.arange(-1, n + 1)
        gradient = np.vstack((gradient, gradient))
        cmap = class_cmap(style, n)
        ax.imshow(gradient, vmin=-0.5, vmax=n - 0.5, cmap=cmap)
        ax.set(xticks=[], yticks=[])



import os
import inspect
from shutil import copyfile

import numpy as np
from astropy.io import fits
from scipy.io import readsav

__all__ = ['make_iter', 'load_parameter', 'merge_results', 'update_signature']


def make_iter(*args):
    """Returns each inputted argument, wrapping in a list if not already iterable.

    Parameters
    ----------
    *args
        Arguments to make iterable.

    Returns
    -------
    iterables
        `*args` converted to iterables.

    Examples
    --------
    >>> make_iter(1)
    [[1]]

    >>> make_iter(1, 2, 3)
    [[1], [2], [3]]

    >>> make_iter(1, [2], 3)
    [[1], [2], [3]]

    It is intended that a list of arguments be passed to the function for conversion:

    >>> make_iter(*[1, [2], 3])
    [[1], [2], [3]]

    Remember that strings are already iterable!

    >>> make_iter(*[[1, 2, 3], (4, 5, 6), "a"])
    [[1, 2, 3], (4, 5, 6), 'a']
    """
    iterables = []  # Holder to return
    for parameter in args:
        try:  # Will work if iterable
            _ = (i for i in parameter)
        except TypeError:  # Not iterable
            parameter = [parameter]  # Wrap in a list
        iterables = iterables + [parameter]  # Append to list to return
    return iterables


def load_parameter(parameter, wl=None):
    """Load parameters from file, optionally evaluating variables from strings.

    Loads the parameter from string or file.

    Parameters
    ----------
    parameter : str
        Parameter to load, either string of Python list/number or filename string. Supported filename extensions are
        '.fits', '.fit', '.fts', '.csv', '.txt', '.npy', '.npz', and '.sav'. If the file does not exist, it will assume
        the string is a Python expression.
    wl : float, optional, default=None
        Central line core wavelength to replace 'wl' in strings. Will only replace occurrences in the `parameter`
        variable itself or in files with extension ".csv" or ".txt". When using `wl`, also use 'inf' and 'nan' as
        required.

    Returns
    -------
    value : numpy.ndarray or list of floats
        Value of parameter in easily computable format (not string).

    Examples
    --------
    >>> load_parameter("wl + 4.2", wl=7.1)
    11.3

    >>> load_parameter("[wl + 4.2, 5.2 - inf, 5 > 3]", wl=7.1)
    [11.3, -inf, 1.0]

    Filenames are given as follows:

    >>> x = load_parameter("datafile.csv", wl=12.4)  # doctest: +SKIP

    >>> x = load_parameter("datafile.fits")  # doctest: +SKIP

    If the file does not exist, the function will assume that the string is a Python expression, possibly leading to an
    error:

    >>> load_parameter("nonexistant.csv")
    Traceback (most recent call last):
     ...
    TypeError: 'NoneType' object is not subscriptable
    """
    if os.path.exists(parameter):  # If the parameter is a real file

        ext = os.path.splitext(parameter)[1]  # File extension

        if ext.lower() in ['.fits', '.fit', '.fts']:  # Extension suggests FITS file

            # Read data from the primary HDU of the FITS file
            hdul = fits.open(parameter)  # Open with mmap
            value = hdul[0].data.copy()  # Copy out of mmap
            hdul.close()  # Close the file

        elif ext.lower() in ['.csv', '.txt']:  # Extension suggests CSV file

            # Read CSV file (assumes a ',' delimiter)
            if wl is not None:  # If `wl` is specified, try a replacement
                value = str(list(np.loadtxt(parameter, delimiter=',', dtype=object))).replace('\'', '')
                try:
                    value = eval(str(value), {'__builtins__': None},
                                 {'wl': wl, 'inf': float('inf'), 'nan': float('nan')})
                except TypeError:  # Only allowed to process `wl` and `inf` variables for security reasons
                    raise SyntaxError("parameter string contains illegal variables")
                except SyntaxError:
                    raise SyntaxError("parameter string '{}' contains a syntax error".format(parameter))
            else:
                value = np.loadtxt(parameter, delimiter=',', dtype=float)

        elif ext.lower() in ['.npy', '.npz']:  # Extension suggests NumPy array

            value = np.load(parameter)

        elif ext.lower() in ['.sav']:  # Extension suggests IDL SAVE file (assumes relevant data in first variable)

            value = list(readsav(parameter).values())[0]

        else:  # Extension not matched
            raise ValueError("loaded parameters can only have file extensions: '.fits', '.fit', '.fts', '.csv', "
                             "'.txt', '.npy', '.npz', '.sav', got '%s'" % ext.lower())

    else:  # Must not be a file (or the filename is incorrect!)

        # Convert to list, calculate relative to central line core (`wl`)
        value = eval(str(parameter), {'__builtins__': None}, {'wl': wl, 'inf': float('inf')})
        try:
            value = [float(val) for val in value]  # Make sure all values are floats
        except TypeError:
            value = float(value)

    return value


def merge_results(filenames, output):
    """Merges files generated by the :meth:`mcalf.models.FitResults.save` method.

    Parameters
    ----------
    filenames : list of str, length>1
        List of FITS files generated by :meth:`mcalf.models.FitResults.save` method.
    output : str
        Name of FITS file to save merged input files to. Will be clobbered.

    Notes
    -----
    See :meth:`mcalf.models.FitResults` for details on the output FITS file data structure.
    """
    if not isinstance(filenames, list) or len(filenames) <= 1:
        raise TypeError("`filenames` must be a list of length greater than 1.")

    # Verification headers (initialise and give keys)
    verification = {
        'PRIMARY': {
            'NTIME': None,
            'NROWS': None,
            'NCOLS': None,
            'TIME': None,
        },
        'PARAMETERS': {
            'NPARAMS': None,
        },
        'CLASSIFICATIONS': {
        },
        'PROFILE': {
            'PROFILES': None
        },
        'SUCCESS': {
        },
        'CHI2': {
        },
        'VLOSA': {
            'VTYPE': None,
            'UNIT': None,
        },
        'VLOSQ': {
            'VTYPE': None,
            'UNIT': None,
        },
    }

    # Values if not fitted (or unsuccessful)
    unset_value = {
        'PRIMARY': '__SKIP__',
        'PARAMETERS': np.nan,
        'CLASSIFICATIONS': -1,
        'PROFILE': 0,
        'SUCCESS': False,
        'CHI2': np.nan,
        'VLOSA': np.nan,
        'VLOSQ': np.nan,
    }

    # Open the output file for updating
    main_hdul = fits.open(filenames[0], mode='readonly')

    # Record the order for easy access {'NAME': index, ...}
    main_index = {main_hdul[v].name: v for v in range(len(main_hdul))}

    # Remove optional keys if not present in first file
    for optional_key in ['VLOSA', 'VLOSQ']:
        if optional_key not in main_index.keys():
            verification.pop(optional_key)

    # Check that the expected HDUs are present
    if main_index.keys() != verification.keys():
        raise ValueError(f"Unexpected HDU name in {filenames[0]}.")

    # Get expected values for the headers from the first file
    for name in verification.keys():
        for attribute in verification[name].keys():
            verification[name][attribute] = main_hdul[main_index[name]].header[attribute]

    # Load the initial arrays
    arrays = {name: main_hdul[main_index[name]].data.copy() for name in verification.keys()}

    # Close the first input file
    main_hdul.close()

    # Copy across the remainder of the FITS files
    for filename in filenames[1:]:
        with fits.open(filename, mode='readonly') as hdul:

            # Check that the expected HDUs are present in `filename`
            input_index = {hdul[v].name: v for v in range(len(hdul))}
            if input_index.keys() != verification.keys():
                raise ValueError(f"Unexpected HDUs in {filename}.")

            for name in verification.keys():  # Loop through the HDUs

                # Verify that the important header items match
                for attribute, expected_value in verification[name].items():
                    if hdul[input_index[name]].header[attribute] != expected_value:
                        # TODO: Handle the case where there are different profiles in each file
                        raise ValueError(f"FITS attribute {attribute} for {name} HDU in {filename} is different.")

                # Create aliases for the input and output arrays
                output_array = arrays[name]
                input_array = hdul[input_index[name]].data

                # Choose the function to test if data is being overwritten
                invalid = unset_value[name]
                if invalid == '__SKIP__':  # PRIMARY HDU (do nothing)
                    continue
                elif np.isnan(invalid):  # floats (can only overwrite nan)
                    test_function = _nan_test
                elif isinstance(invalid, bool) and not invalid:  # bool (can only overwrite False)
                    test_function = _false_test
                elif isinstance(invalid, (int, np.integer)) and invalid == -1:
                    test_function = _minus_one_test
                elif isinstance(invalid, (int, np.integer)) and invalid == 0:
                    test_function = _zero_test
                else:
                    raise ValueError(f"Unexpected invalid value {invalid}.")

                # Verify that no data is being overwritten
                should_edit = test_function(input_array)
                would_edit = output_array[should_edit]
                if np.sum(test_function(would_edit)) != 0:
                    raise ValueError(f"Overlapping values in {name} HDU at {filename}.")

                # Merge `input_array` onto output
                output_array[np.where(should_edit)] = input_array[np.where(should_edit)]

    # Copy the first FITS input to the output file
    copyfile(filenames[0], output)

    # Open the output file for updating
    with fits.open(output, mode='update') as output_hdul:
        for hdu in output_hdul:
            hdu.data = arrays[hdu.name]


def _nan_test(x):
    """Finds where not NaN.

    False if index is NaN.

    Parameters
    ----------
    x : array_like
        Array to search.

    Returns
    -------
    array : array of bool
        Whether corresponding index is not NaN.
    """
    return ~np.isnan(x)


def _false_test(x):
    """Finds where not False (where is True).

    Parameters
    ----------
    x : array_like
        Array to search.

    Returns
    -------
    array : array of bool
        Whether corresponding index is True. (Is not False.)

    Notes
    -----
    Converts to bool dtype as integer could have been given.
    """
    return x.astype(bool)


def _minus_one_test(x):
    """Finds where not -1.

    Parameters
    ----------
    x : array_like
        Array to search.

    Returns
    -------
    array : array of bool
        Whether corresponding index is not -1.
    """
    return x != -1


def _zero_test(x):
    """Finds where not 0.

    Parameters
    ----------
    x : array_like
        Array to search.

    Returns
    -------
    array : array of bool
        Whether corresponding index is not 0.
    """
    return x != 0


def _as_keywords(dictionary):
    """Converts a (ordered) dictionary of name, default value pairs into a
    list of KEYWORD_ONLY Parameters."""
    return [
        inspect.Parameter(name, inspect.Parameter.KEYWORD_ONLY, default=default)
        for name, default in dictionary.items()
    ]


def _filter_params(parameters):
    """Filters a list of Parameters such that all are converted to
    KEYWORD_ONLY and *args, **kwargs and self are removed."""
    return [
        param.replace(kind=inspect.Parameter.KEYWORD_ONLY) for param in parameters
        if (param.kind not in (
            inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD
        )) and (param.name != 'self')
    ]


def _drop_duplicate_names(params: list):
    """Drops duplicate Parameters from a list."""
    existing = []
    drop = []
    for i in range(len(params)-1, -1, -1):
        if params[i].name in existing:
            drop.append(i)
        else:
            existing.append(params[i].name)
    for i in drop[::-1]:
        params.pop(i)


def _update_parameters(params, cls, parse_defaults=True):
    """Update a list of Parameters.

    Parameters
    ----------
    params : list[`inspect.Parameter`]
        List of Parameters to update (inplace).
    cls : type
        Class to extract `__init__` signature from.
    parse_defaults : bool, optional, default=True
        Whether to include Parameters from `cls.default_kwargs`.

    Returns
    -------
    sig : `inspect.Signature`
        Signature of `cls.__init__`.
    """
    if parse_defaults:
        params += _as_keywords(cls.default_kwargs)
    sig = inspect.signature(cls.__init__)
    params += _filter_params(sig.parameters.values())
    _drop_duplicate_names(params)
    return sig


def update_signature(cls):
    """Update the signature of a model class.

    Parameters
    ----------
    cls : type
        The model class to set a `cls.__init__.__signature__` for.

    Notes
    -----
    This should be called during import of the model class.
    This function should be called for every class in the
    model class hierarchy in order starting from
    `~mcalf.models.ModelBase`.
    """
    params = []

    all_classes = inspect.getmro(cls)
    if len(all_classes) >= 3:  # ([`cls`, [...,]] `ModelBase`, `object`)
        _update_parameters(params, all_classes[1], parse_defaults=False)

    sig = _update_parameters(params, cls)
    new_sig = sig.replace(parameters=params)
    cls.__init__.__signature__ = new_sig



import os
import copy
import warnings
import collections

import numpy as np
from pathos.multiprocessing import ProcessPool as Pool
from scipy.optimize import curve_fit
from sklearn.exceptions import NotFittedError
from sklearn.metrics import classification_report
from yaml import load

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from mcalf.utils.misc import load_parameter, make_iter, update_signature
from mcalf.utils.spec import reinterpolate_spectrum

__all__ = ['ModelBase', 'BASE_PARAMETERS', 'BASE_ATTRIBUTES']


class ModelBase:
    """Base class for spectral line model fitting.

    .. warning::
        This class should not be used directly.
        Use derived classes instead.

    Parameters
    ----------
    ${PARAMETERS}

    Attributes
    ----------
    ${ATTRIBUTES}
    """

    default_kwargs = default_modelbase_kwargs = collections.OrderedDict([
        ('stationary_line_core', None),
        ('original_wavelengths', None),
        ('constant_wavelengths', None),
        ('delta_lambda', 0.05),
        ('sigma', None),
        ('prefilter_response', None),
        ('prefilter_ref_main', None),
        ('prefilter_ref_wvscl', None),
        ('output', None),
    ])

    def __init__(self, config=None, **kwargs):

        # STAGE 1: Define dictionary of default attribute values
        defaults = copy.deepcopy(self.default_modelbase_kwargs)

        # STAGE 2: Update defaults with any values specified in a config file
        self.config = {}  # Dictionary of parameters from the config file
        if config is not None:  # Process config file if one is specified

            with open(config, 'r') as stream:  # Load YAML file
                self.config = load(stream, Loader=Loader)

            # Extract the BaseModel attributes from the config file
            class_defaults = {k: self.config[k] for k in self.config.keys() if k in defaults.keys()}
            for k in class_defaults.keys():
                if k not in ['output']:  # Keep string as is (don't pass through `load_parameter`)
                    class_defaults[k] = load_parameter(class_defaults[k])
                self.config.pop(k)  # Remove copied parameter

            # Update the defaults with the config file
            defaults.update(class_defaults)

        # STAGE 3: Update defaults with the keyword arguments passed into the class initialisation
        # Verify that only expected kwargs are passed
        for k in kwargs.keys():
            if k not in defaults.keys():
                raise TypeError(f"ModelBase() got an unexpected keyword argument '{k}'")
        defaults.update(kwargs)  # Update the defaults

        # Load default values of any parameters that haven't been given yet
        if defaults['original_wavelengths'] is None:
            raise ValueError("original_wavelengths must be specified")
        if defaults['constant_wavelengths'] is None:
            defaults['constant_wavelengths'] = np.arange(min(defaults['original_wavelengths']),
                                                         (max(defaults['original_wavelengths'])
                                                         + defaults['delta_lambda']),
                                                         defaults['delta_lambda'])

        # STAGE 4: Set the object attributes (with some type enforcing)
        # values in the defaults dict
        self.original_wavelengths = np.asarray(defaults['original_wavelengths'], dtype=np.float64)
        self.constant_wavelengths = np.asarray(defaults['constant_wavelengths'], dtype=np.float64)
        if defaults['stationary_line_core'] is None:  # Allow to be None by default
            self.__stationary_line_core = None  # Override setter
        else:
            self.stationary_line_core = defaults['stationary_line_core']
        self.__delta_lambda = defaults['delta_lambda']
        self.sigma = defaults['sigma']
        self.prefilter_response = defaults['prefilter_response']
        self.__prefilter_ref_main = defaults['prefilter_ref_main']
        self.__prefilter_ref_wvscl = defaults['prefilter_ref_wvscl']
        self.output = defaults['output']
        # attributes whose default value cannot be changed during initialisation
        self.neural_network = None  # Must be set in a child class initialisation or after initialisation
        self.array = None  # Array that will hold any loaded spectra
        self.background = None  # Array holding constant background values for `self.array`

        # STAGE 5: Validate the loaded attributes
        self._validate_base_attributes()

    @property
    def stationary_line_core(self):
        return self.__stationary_line_core

    @stationary_line_core.setter
    def stationary_line_core(self, wavelength):

        # Verify that stationary_line_core is a float
        if not isinstance(wavelength, float):
            raise ValueError("stationary_line_core must be a float, got %s" % type(wavelength))

        # Stationary wavelength must be within wavelength range
        original_diff = self.original_wavelengths - wavelength
        constant_diff = self.constant_wavelengths - wavelength
        for n, i in [['original_wavelengths', original_diff], ['constant_wavelengths', constant_diff]]:
            if min(i) > 1e-6 or max(i) < -1e-6:
                raise ValueError("`stationary_line_core` is not within `{}`".format(n))

        # Verification passed; set value
        self.__stationary_line_core = wavelength

    def _set_prefilter(self):
        """Set the `prefilter_response` parameter.

        .. deprecated:: 0.2
             Prefilter response correction code, and `prefilter_response`, `prefilter_ref_main`
             and `prefilter_ref_wvscl`, may be removed in a later release of MCALF.
             Spectra should be fully processed before loading into MCALF.

        This method should be called in a child class once `stationary_line_core` has been set.
        """
        if self.prefilter_response is None:
            if self.__prefilter_ref_main is not None and self.__prefilter_ref_wvscl is not None:
                self.prefilter_response = reinterpolate_spectrum(self.__prefilter_ref_main,
                                                                 self.__prefilter_ref_wvscl + self.stationary_line_core,
                                                                 self.constant_wavelengths)
            else:
                return  # None of the prefilter attributes are set, so no prefilter to apply.
        else:  # Make sure it is a numpy array so that division works as expected when doing array operations
            self.prefilter_response = np.asarray(self.prefilter_response, dtype=np.float64)

        # TODO: Remove this warning and possibly all prefilter code
        warnings.warn("Spectra should be fully processed before loading into MCALF. "
                      "Prefilter response correction code may be removed in a later "
                      "release.", PendingDeprecationWarning)

        # If a prefilter response is given it must be a compatible length
        if self.prefilter_response is not None:
            if len(self.prefilter_response) != len(self.constant_wavelengths):
                raise ValueError("prefilter_response array must be the same length as constant_wavelengths array")

    def _validate_base_attributes(self):
        """Validate some of the object's attributes.

        Raises
        ------
        ValueError
            To signal that an attribute is not valid.
        """
        # Wavelength arrays must be sorted ascending
        if np.sum(np.diff(self.original_wavelengths) > 0) < len(self.original_wavelengths) - 1:
            raise ValueError("original_wavelengths array must be sorted ascending")
        if np.sum(np.diff(self.constant_wavelengths) > 0) < len(self.constant_wavelengths) - 1:
            raise ValueError("constant_wavelengths array must be sorted ascending")

        # Warn if the constant wavelengths extrapolate the original wavelengths
        if min(self.constant_wavelengths) - min(self.original_wavelengths) < -1e-6:
            # If lower-bound of constant wavelengths is more than 1e-6 outside of the original wavelengths
            warnings.warn("Lower bound of `constant_wavelengths` is outside of `original_wavelengths` range.")
        if max(self.constant_wavelengths) - max(self.original_wavelengths) - self.__delta_lambda > 1e-6:
            # If upper-bound of constant wavelengths is more than 1e-6 ouside the original wavelengths
            warnings.warn("Upper bound of `constant_wavelengths` is outside of `original_wavelengths` range.")

    def _load_data(self, array, names=None, target=None):
        """Load a specified array into the model object.

        Load `array` with dimension names `names` into the attribute specified by `target`.

        Parameters
        ----------
        array : numpy.ndarray
            The array to load.

        names : list of str, length=`array.ndim`
            List of dimension names for `array`. Valid dimension names depend on `target`.

        target : {'array', 'background'}
            The attribute to load the `array` into.

        See Also
        --------
        load_array : Load and array of spectra.
        load_background : Load an array of spectral backgrounds.
        """

        # Validate specified `target`
        if target not in ['array', 'background']:
            raise ValueError("array target must be either 'array' or 'background', got '%s'" % target)

        # Verify `array` has data on more than one spectrum
        if (target == 'array' and np.ndim(array) == 1) or (target == 'background' and np.ndim(array) == 0):
            raise ValueError("cannot load an array containing one spectrum, use the spectrum directly instead")

        # Validate the dimension names given
        if names is None:
            raise ValueError("dimension names must be specified for each dimension")
        if np.ndim(array) != len(np.atleast_1d(names)):  # All dimensions must be labelled
            raise ValueError("number of dimension names do not match number of columns")
        if len(np.atleast_1d(names)) != len(np.unique(np.atleast_1d(names))):
            raise ValueError("duplicate dimension names found")
        if target == 'array' and 'wavelength' not in names:
            raise ValueError("array must contain a wavelength dimension")

        # # Pad input array to include all possible input dimensions and
        # #     then transpose it such that order of dimensions is consistent
        #   e.g. ['time', 'wavelength']           -> ['time', 'row', 'column', 'wavelength']
        #        ['wavelength', 'column', 'row']  -> ['time', 'row', 'column', 'wavelength']

        # Define the transposition on the input array to get the dimensions in order
        #     ['time', 'row', 'column'(, 'wavelength')]
        n_dims = 4 if target == 'array' else 3  # ndim of (padded) array to load data into
        transposition = [None] * n_dims
        for i in range(len(names)):
            if names[i] == 'time':
                transposition[0] = i
            elif names[i] == 'row':
                transposition[1] = i
            elif names[i] == 'column':
                transposition[2] = i
            elif names[i] == 'wavelength' and target == 'array':  # Only `target` 'array' should contain wavelengths
                transposition[3] = i
            else:  # Name given is not recognised
                raise ValueError("name '{}' is not a valid dimension name".format(names[i]))

        # Define where the new padded dimensions will be placed within the transposition
        array_n_dims = np.ndim(array)  # Number of dimensions in input array
        new_axis_id = array_n_dims  # Start labelling new axes after last dimension of input array
        for axis in range(n_dims):  # Loop through each axis of padded array to be created
            if transposition[axis] is None:  # Axis hasn't been labelled yet (not part of input array)
                transposition[axis] = new_axis_id  # Locate it within the transposition definition
                new_axis_id += 1

        # Add missing dimensions to the end
        n_dims_to_add = n_dims - array_n_dims
        ordered_array = array
        for i in range(n_dims_to_add):  # Add required number of dimensions to the end of the input array
            ordered_array = ordered_array[..., np.newaxis]

        # Reorder the dimensions of the padded array such that they can be later assumed
        ordered_array = np.transpose(ordered_array, transposition)

        # If spectral array, verify that length of each spectrum is equal to the number of corresponding wavelengths
        if target == 'array' and np.shape(ordered_array)[-1] != len(self.original_wavelengths):
            raise ValueError("length of wavelength dimension not equal length of original_wavelengths")

        # Load the processed array into the specified object parameter
        if target == 'array':
            self.array = ordered_array
        elif target == 'background':
            self.background = ordered_array

        # If "the other array" has been loaded, warn the user if their dimensions are not compatible
        if self.array is not None and self.background is not None and \
                np.sum(np.shape(self.array)[:-1] != np.shape(self.background)) != 0:
            warnings.simplefilter('always', UserWarning)
            warnings.warn("shape of spectrum array indices does not match shape of background array")

    def load_array(self, array, names=None):
        """Load an array of spectra.

        Load `array` with dimension names `names` into the `array` parameter of the model object.

        Parameters
        ----------
        array : numpy.ndarray, ndim>1
            An array containing at least two spectra.

        names : list of str, length=`array.ndim`
            List of dimension names for `array`. Valid dimension names are 'time', 'row', 'column' and 'wavelength'.
            'wavelength' is a required dimension.

        See Also
        --------
        load_background : Load an array of spectral backgrounds.

        Examples
        --------

        Create a basic model:

        >>> import mcalf.models
        >>> from astropy.io import fits
        >>> wavelengths = [0.0, 10.0, 20.0, 30.0, 40.0, 50.0]
        >>> model = mcalf.models.ModelBase(original_wavelengths=wavelengths)

        Load spectra from a file:

        >>> spectra = fits.open('spectra_0000.fits')[0].data  # doctest: +SKIP
        >>> model.load_array(spectra, names=['wavelength', 'column', 'row'])  # doctest: +SKIP
        """
        self._load_data(array, names=names, target='array')

    def load_background(self, array, names=None):
        """Load an array of spectral backgrounds.

        Load `array` with dimension names `names` into `background` parameter of the model object.

        Parameters
        ----------
        array : numpy.ndarray, ndim>0
            An array containing at least two backgrounds.

        names : list of str, length=`array.ndim`
            List of dimension names for `array`. Valid dimension names are 'time', 'row' and 'column'.

        See Also
        --------
        load_array : Load and array of spectra.

        Examples
        --------

        Create a basic model:

        >>> import mcalf.models
        >>> from astropy.io import fits
        >>> wavelengths = [0.0, 10.0, 20.0, 30.0, 40.0, 50.0]
        >>> model = mcalf.models.ModelBase(original_wavelengths=wavelengths)

        Load background array from a file:

        >>> background = fits.open('background_0000.fits')[0].data  # doctest: +SKIP
        >>> model.load_background(background, names=['column', 'row'])  # doctest: +SKIP
        """
        self._load_data(array, names=names, target='background')

    def train(self, X, y):
        """Fit the neural network model to spectra matrix X and spectra labels y.

        Calls the :meth:`fit` method on the `neural_network` parameter of the model object.

        Parameters
        ----------
        X : numpy.ndarray or sparse matrix, shape=(n_spectra, n_wavelengths)
            The input spectra.
        y : numpy.ndarray, shape= (n_spectra,) or (n_spectra, n_outputs)
            The target class labels.

        See Also
        --------
        test : Test how well the neural network has been trained.
        """
        self.neural_network.fit(X, y)

    def test(self, X, y):
        """Test the accuracy of the trained neural network.

        Prints a table of results showing:

        1) the percentage of predictions that equal the target labels;
        2) the average classification deviation and standard deviation from the ground truth classification
           for each labelled classification;
        3) the average classification deviation and standard deviation overall.

        If the model object has an output parameter, it will create a CSV file (``output``/neural_network/test.csv)
        listing the predictions and ground truth data.

        Parameters
        ----------
        X : numpy.ndarray or sparse matrix, shape=(n_spectra, n_wavelengths)
            The input spectra.
        y : numpy.ndarray, shape= (n_spectra,) or (n_spectra, n_outputs)
            The target class labels.

        See Also
        --------
        train : Train the neural network.
        """
        # Predict the labels
        try:
            predictions = self.neural_network.predict(X)
        except NotFittedError:
            raise NotFittedError("Neural network has not been trained yet. Call 'train' on the model class first.")
        if len(predictions) != len(y):  # Must be a corresponding ground truth label to compare to
            raise ValueError("number of samples X does not equal number of labels y")

        # Score is the percentage of predictions that are equal to their ground truth label
        score = np.sum(predictions == y) / len(predictions) * 100

        # Generate the table
        info = "+---------------------------------------------+\n"
        info += "|      Neural Network Testing Statistics      |\n"
        info += "+---------------------------------------------+\n"
        info += "| Percentage predictions==labels :: {: 6.2f}%   |\n".format(score)
        info += "+---------------------------------------------+\n"
        info += "| Average deviation for each classification   |\n"
        info += "+---------------------------------------------+\n"

        # Generate the average deviation for each classification
        classifications = np.sort(np.unique([predictions, y]))
        for classification in classifications:
            dev = predictions[np.where(y == classification)[0]] - classification
            info += "|     class {: 2d} :: {: 6.2f} ± {: 4.2f}              |\n".format(classification, np.mean(dev),
                                                                                        np.std(dev))
            info += "+---------------------------------------------+\n"

        # Generate the average deviation over all classifications
        dev = predictions - y
        info += "| Average deviation overall :: {: 6.2f} ± {: 4.2f} |\n".format(np.mean(dev), np.std(dev))
        info += "+---------------------------------------------+"

        # Print the table
        print(info)

        # Print scikit-learn classification report
        print(classification_report(y, predictions))

        # Output to file columns: 'ground_truth', 'predictions'
        if self.output is not None:
            try:
                os.makedirs(os.path.join(self.output, "neural_network"))
            except FileExistsError:
                pass
            np.savetxt(os.path.join(self.output, "neural_network", "test.csv"), np.asarray([y, predictions]).T,
                       fmt='%3d', delimiter=',', header='ground_truth, prediction')

    def classify_spectra(self, time=None, row=None, column=None, spectra=None, only_normalise=False):
        """Classify the specified spectra.

        Will also normalise each spectrum such that its intensity will range from zero to one.

        Parameters
        ----------
        time : int or iterable, optional, default=None
            The time index. The index can be either a single integer index or an iterable. E.g. a list,
            a :class:`numpy.ndarray`, a Python range, etc. can be used.
        row : int or iterable, optional, default=None
            The row index. See comment for `time` parameter.
        column : int or iterable, optional, default=None
            The column index. See comment for `time` parameter.
        spectra : numpy.ndarray, optional, default=None
            The explicit spectra to classify. If `only_normalise` is False, this must be 1D.
            However, if `only_normalise` is set to true, `spectra` can be of any dimension.
            It is assumed that the final dimension is wavelengths, so return shape will be the
            same as `spectra`, except with no final wavelengths dimension.
        only_normalise : bool, optional, default=False
            Whether the single spectrum given  in `spectra` should not be interpolated and corrected.
            If set to true, the only processing applied to `spectra` will be a normalisation to be
            in range 0 to 1.

        Returns
        -------
        classifications : numpy.ndarray
            Array of classifications with the same time, row and column indices as `spectra`.

        See Also
        --------
        train : Train the neural network.
        test : Test the accuracy of the neural network.
        get_spectra : Get processed spectra from the objects `array` attribute.

        Examples
        --------

        Create a basic model:

        >>> import mcalf.models
        >>> import numpy as np
        >>> wavelengths = np.linspace(8542.1, 8542.2, 30)
        >>> model = mcalf.models.ModelBase(original_wavelengths=wavelengths)

        Load a trained neural network:

        >>> import pickle
        >>> pkl = open('trained_neural_network.pkl', 'rb')  # doctest: +SKIP
        >>> model.neural_network = pickle.load(pkl)  # doctest: +SKIP

        Classify an individual spectrum:

        >>> spectrum = np.random.rand(30)
        >>> model.classify_spectra(spectra=spectrum)  # doctest: +SKIP
        array([2])

        When :code:`only_normalise=True`, classify an n-dimensional spectral array:

        >>> spectra = np.random.rand(5, 4, 3, 2, 30)
        >>> model.classify_spectra(spectra=spectra, only_normalise=True).shape  # doctest: +SKIP
        (5, 4, 3, 2)

        Load spectra from a file and classify:

        >>> from astropy.io import fits
        >>> spectra = fits.open('spectra_0000.fits')[0].data  # doctest: +SKIP
        >>> model.load_array(spectra, names=['wavelength', 'column', 'row'])  # doctest: +SKIP
        >>> model.classify_spectra(column=range(10, 15), row=[7, 16])  # doctest: +SKIP
        array([[[0, 2, 0, 3, 0],
                [4, 0, 1, 0, 0]]])
        """
        if not only_normalise:  # Get the spectrum, otherwise use the provided one directly
            spectra = self.get_spectra(time=time, row=row, column=column, spectrum=spectra)

        # Vectorised normalisation for all spectra such that intensities for each spectrum is in range [0, 1]
        spectra_list = spectra.reshape(-1, spectra.shape[-1]).T  # Reshape & transpose to (n_wavelengths, n_spectra)
        spectra_list = spectra_list - spectra_list.min(axis=0)  # Subtract each spectrum's min value from itself (min 0)
        spectra_list = spectra_list / spectra_list.max(axis=0)  # Divide each spectrum by its max value (min 0, max 1)
        spectra_list = spectra_list.T  # Transpose to (n_spectra, n_wavelengths)

        # Create the empty classifications array
        classifications = np.full(len(spectra_list), -1, dtype=int)  # -1 reserved for invalid
        valid_spectra_i = np.where(~np.isnan(spectra_list[:, 0]))  # Only predict the valid spectra
        # Note: Spectra that have indices in the data but were masked or outside the field of view should be np.nan
        try:
            classifications[valid_spectra_i] = self.neural_network.predict(spectra_list[valid_spectra_i])
        except NotFittedError:
            raise NotFittedError("Neural network has not been trained yet. Call 'train' on the model class first.")
        try:  # Try to give the classifications array the same indices as the spectral array
            classifications = classifications.reshape(*spectra.shape[:-1])
        except TypeError:  # 1D spectra array given
            assert classifications.size == 1  # Verify

        return classifications

    def _get_time_row_column(self, time=None, row=None, column=None):
        """Validate and infer the time, row and column index.

        Takes any time, row and column index given and if any are not specified, they are returned as 0 if the
        spectral array only has one value at its dimension. If there are multiple and no index is specified,
        an error is raised due to the ambiguity.

        Parameters
        ----------
        time : optional, default=None
            The time index.
        row : optional, default=None
            The row index.
        column : optional, default=None
            The column index.

        Returns
        -------
        time
            The corrected time index.
        row
            The corrected row index.
        column
            The corrected column index.

        See Also
        --------
        mcalf.utils.misc.make_iter : Make a variable iterable.

        Notes
        -----
        No type checking is done on the input indices so it can be anything but in most cases will need to be
        either an integer or iterable. The :func:`mcalf.utils.misc.make_iter` function can be used to make
        indices iterable.
        """
        array_shape = np.shape(self.array)
        if time is None:
            # No index was specified but we can infer that it would be 0 as there is only 1 value at this dimension
            if array_shape[0] == 1:
                time = 0
            else:
                raise ValueError("time index must be specified as multiple indices exist")
        if row is None:
            if array_shape[1] == 1:
                row = 0
            else:
                raise ValueError("row index must be specified as multiple indices exist")
        if column is None:
            if array_shape[2] == 1:
                column = 0
            else:
                raise ValueError("column index must be specified as multiple indices exist")
        return time, row, column

    def get_spectra(self, time=None, row=None, column=None, spectrum=None, correct=True, background=False):
        """Gets corrected spectra from the spectral array.

        Takes either a set of indices or an explicit spectrum and optionally applied corrections and background
        removal.

        Parameters
        ----------
        time : int or iterable, optional, default=None
            The time index. The index can be either a single integer index or an iterable. E.g. a list,
            a :class:`numpy.ndarray`, a Python range, etc. can be used.
        row : int or iterable, optional, default=None
            The row index. See comment for `time` parameter.
        column : int or iterable, optional, default=None
            The column index. See comment for `time` parameter.
        spectrum : ndarray of ndim=1, optional, default=None
            The explicit spectrum. If provided, `time`, `row`, and `column` are ignored.
        correct : bool, optional, default=True
            Whether to reinterpolate the spectrum and apply the prefilter correction (if exists).
        background : bool, optional, default=False
            Whether to include the background in the outputted spectra. Only removes the background if the
            relevant background array has been loaded. Does not remove background is processing an explicit spectrum.

        Returns
        -------
        spectra : ndarray

        Examples
        --------

        Create a basic model:

        >>> import mcalf.models
        >>> import numpy as np
        >>> wavelengths = np.linspace(8541.3, 8542.7, 30)
        >>> model = mcalf.models.ModelBase(original_wavelengths=wavelengths)

        Provide a single spectrum for processing, and notice output is 1D:

        >>> spectrum = model.get_spectra(spectrum=np.random.rand(30))
        >>> spectrum.ndim
        1

        Load an array of spectra:

        >>> spectra = np.random.rand(3, 4, 30)
        >>> model.load_array(spectra, names=['column', 'row', 'wavelength'])

        Extract a single (unprocessed) spectrum from the loaded array, and notice output is 4D:

        >>> spectrum = model.get_spectra(row=1, column=0, correct=False)
        >>> spectrum.shape
        (1, 1, 1, 30)
        >>> (spectrum[0, 0, 0] == spectra[0, 1]).all()
        True

        Extract an array of spectra, and notice output is 4D, and with dimensions
        time, row, column, wavelength regardless of the original dimensions and order:

        >>> spectrum = model.get_spectra(row=range(4), column=range(3))
        >>> spectrum.shape
        (1, 4, 3, 30)

        Notice that the time index can be excluded, as the loaded array only represents a single time.
        However, in this case leaving out `row` or `column` results in an error as it is ambiguous:

        >>> spectrum = model.get_spectra(row=range(4))
        Traceback (most recent call last):
         ...
        ValueError: column index must be specified as multiple indices exist
        """
        # Locate the spectra
        if spectrum is None:  # No explicit spectrum so use specified indices
            time, row, column = make_iter(*self._get_time_row_column(time=time, row=row, column=column))
            spectra = self.array[time][:, row][:, :, column]  # Iterable indices are needed to crop array correctly
        else:
            if np.ndim(spectrum) != 1:  # Keep things simple (for now)
                raise ValueError("explicit spectrum must have one dimension, got %s" % np.ndim(spectrum))
            spectra = spectrum

        # Apply option corrections to spectra
        if correct:
            spectra_list = spectra.reshape(-1, spectra.shape[-1])  # To shape: (n_spectra, n_wavelengths)
            corrected_spectra = np.array([
                reinterpolate_spectrum(spectrum, self.original_wavelengths, self.constant_wavelengths)
                for spectrum in spectra_list])  # Reinterpolate each spectrum

            if self.prefilter_response is not None:  # Apply prefilter correction if one was initialised
                corrected_spectra /= self.prefilter_response

            spectra = corrected_spectra.reshape(*spectra.shape[:-1], -1)  # Revert to the original shape

        # Apply optional background removal to spectra
        if spectrum is None and not background and self.background is not None:  # Remove the background
            # Calculate transpositions to get the array in the right shape for a quick background subtraction
            forward_transpose = list(np.roll(np.arange(spectra.ndim), 1))
            backwards_transpose = list(np.roll(np.arange(spectra.ndim), -1))
            spectra = np.transpose(spectra, axes=forward_transpose) - self.background[time][:, row][:, :, column]
            spectra = np.transpose(spectra, axes=backwards_transpose)  # Revert to original shape

        return spectra

    def _fit(self, spectrum, classification=None, spectrum_index=None):
        """Fit a single spectrum for the given profile or classification.

        .. warning::
            This call signature and docstring specify how the `_fit` method must be implemented in
            each subclass of `ModelBase`. **It is not implemented in this class.**

        Parameters
        ----------
        spectrum : numpy.ndarray, ndim=1, length=n_constant_wavelengths
            The spectrum to be fitted.
        classification : int, optional, default=None
            Classification to determine the fitted profile to use.
        spectrum_index : array_like or list or tuple, length=3, optional, default=None
            The [time, row, column] index of the `spectrum` provided. Only used for error reporting.

        Returns
        -------
        result : mcalf.models.FitResult
            Outcome of the fit returned in a :class:`mcalf.models.FitResult` object.

        See Also
        --------
        fit : The recommended method for fitting spectra.
        mcalf.models.FitResult : The object that the fit method returns.

        Notes
        -----
        This method is called for each requested spectrum by the :meth:`models.ModelBase.fit` method.
        This is where most of the adjustments to the fitting method should be made. See other
        subclasses of `models.ModelBase` for examples of how to implement this method in a
        new subclass. See :meth:`models.ModelBase.fit` for more information on how this method is
        called.
        """
        raise NotImplementedError("The `_fit` method must be implemented in a subclass of `ModelBase`."
                                  "See `models.ModelBase._fit` for more details.")

    def fit(self, time=None, row=None, column=None, spectrum=None, classifications=None,
            background=None, n_pools=None, **kwargs):
        """Fits the model to specified spectra.

        Fits the model to an array of spectra using multiprocessing if requested.

        Parameters
        ----------
        time : int or iterable, optional, default=None
            The time index. The index can be either a single integer index or an iterable. E.g. a list,
            :class:`numpy.ndarray`, a Python range, etc. can be used.
        row : int or iterable, optional, default=None
            The row index. See comment for `time` parameter.
        column : int or iterable, optional, default=None
            The column index. See comment for `time` parameter.
        spectrum : numpy.ndarray, ndim=1, optional, default=None
            The explicit spectrum to fit the model to.
        classifications : int or array_like, optional, default=None
            Classifications to determine the fitted profile to use. Will use neural network to classify them if not.
            If a multidimensional array, must have the same shape as [`time`, `row`, `column`].
            Dimensions that would have length of 1 can be excluded.
        background : float, optional, default=None
            If provided, this value will be subtracted from the explicit spectrum provided in `spectrum`. Will
            not be applied to spectra found from the indices, use the :meth:`~mcalf.models.ModelBase.load_background`
            method instead.
        n_pools : int, optional, default=None
            The number of processing pools to calculate the fitting over. This allocates the fitting of different
            spectra to `n_pools` separate worker processes. When processing a large number of spectra this will make
            the fitting process take less time overall. It also distributes such that each worker process has the
            same ratio of classifications to process. This should balance out the workload between workers.
            If few spectra are being fitted, performance may decrease due to the overhead associated with splitting
            the evaluation over separate processes. If `n_pools` is not an integer greater than zero, it will fit
            the spectrum with a for loop.
        **kwargs : dict, optional
            Extra keyword arguments to pass to :meth:`~mcalf.models.ModelBase._fit`.

        Returns
        -------
        result : list of :class:`~mcalf.models.FitResult`, length=n_spectra
            Outcome of the fits returned as a list of :class:`~mcalf.models.FitResult` objects.

        Examples
        --------

        Create a basic model:

        >>> import mcalf.models
        >>> import numpy as np
        >>> wavelengths = np.linspace(8541.3, 8542.7, 30)
        >>> model = mcalf.models.ModelBase(original_wavelengths=wavelengths)

        Set up the neural network classifier:

        >>> model.neural_network = ...  # load an untrained classifier  # doctest: +SKIP
        >>> model.train(...)  # doctest: +SKIP
        >>> model.test(...)  # doctest: +SKIP

        Load the spectra and background array:

        >>> model.load_array(...)  # doctest: +SKIP
        >>> model.load_background(...)  # doctest: +SKIP

        Fit a subset of the loaded spectra, using 5 processing pools:

        >>> fits = model.fit(row=range(3, 5), column=range(200), n_pools=5)  # doctest: +SKIP
        >>> fits  # doctest: +SKIP
        ['Successful FitResult with ________ profile of classification 0',
         'Successful FitResult with ________ profile of classification 2',
         ...
         'Successful FitResult with ________ profile of classification 0',
         'Successful FitResult with ________ profile of classification 4']

        Merge the fit results into a :class:`~mcalf.models.FitResults` object:

        >>> results = mcalf.models.FitResults((500, 500), 8)
        >>> for fit in fits:  # doctest: +SKIP
        ...     results.append(fit)  # doctest: +SKIP

        See :meth:`fit_spectrum` examples for how to manually providing a `spectrum` to fit.
        """
        # Only include the background is using an explicit spectrum; remove for indices
        include_background = False if spectrum is None else True
        spectra = self.get_spectra(time=time, row=row, column=column, spectrum=spectrum, background=include_background)

        # At least one valid spectrum must be given
        n_valid = np.sum(~np.isnan(spectra[..., 0]))
        if np.sum(~np.isnan(spectra[..., 0])) == 0:
            raise ValueError("no valid spectra given")

        if spectrum is not None and background is not None:  # remove explicit background of explicit spectrum
            spectra -= background

        if classifications is None:  # Classify if not already specified
            classifications = self.classify_spectra(spectra=spectra, only_normalise=True)

        if spectrum is None:  # No explicit spectrum given

            # Create the array of indices that the spectra represent
            time, row, column = make_iter(*self._get_time_row_column(time=time, row=row, column=column))
            indices = np.transpose(np.array(np.meshgrid(time, row, column, indexing='ij')), axes=[1, 2, 3, 0])

            # Ensure that length of arrays match
            spectra_indices_shape_mismatch = spectra.shape[:-1] != indices.shape[:-1]
            spectra_class_size_mismatch = np.size(spectra[..., 0]) != np.size(classifications)
            spectra_class_shape_mismatch = False  # Only test this if appropriate
            if isinstance(classifications, np.ndarray) and classifications.ndim != 1:
                # If a multidimensional array of classifications are given, make sure it matches the indices layout
                # Allow for dimensions of length 1 to be excluded
                if np.squeeze(spectra[..., 0]).shape != np.squeeze(classifications).shape:
                    spectra_class_shape_mismatch = True
            if spectra_indices_shape_mismatch or spectra_class_size_mismatch or spectra_class_shape_mismatch:
                raise ValueError("number classifications do not match number of spectra and associated indices")

            # Make shape (n_spectra, n_features) so can process in a list
            spectra = spectra.reshape(-1, spectra.shape[-1])
            indices = indices.reshape(-1, indices.shape[-1])
            classifications = np.asarray(classifications)  # Make sure ndarray methods are available
            classifications = classifications.reshape(-1)

            # Remove spectra that are invalid (this allows for masking of the loaded data to constrain a region to fit)
            valid_spectra_i = np.where(~np.isnan(spectra[:, 0]))  # Where the first item of the spectrum is not NaN
            spectra = spectra[valid_spectra_i]
            indices = indices[valid_spectra_i]
            classifications = classifications[valid_spectra_i]

            assert len(spectra) == len(indices) == len(classifications)  # Postprocessing sanity check

            # Multiprocessing not required
            if n_pools is None or (isinstance(n_pools, (int, np.integer)) and n_pools <= 0):

                print("Processing {} spectra".format(n_valid))
                results = [
                    self._fit(spectra[i], classification=classifications[i], spectrum_index=indices[i], **kwargs)
                    for i in range(len(spectra))
                ]

            elif isinstance(n_pools, (int, np.integer)) and n_pools >= 1:  # Use multiprocessing

                # Define single argument function that can be evaluated in the pools
                def func(data, kwargs=kwargs):
                    # Extract data and pass to `_fit` method
                    spectrum, index, classification = data  # pragma: no cover
                    return self._fit(spectrum, classification=classification,
                                     spectrum_index=list(index), **kwargs)  # pragma: no cover

                # Sort the arrays in descending classification order
                s = np.argsort(classifications)[::-1]  # Classifications indices sorted in descending order
                spectra = spectra[s]
                indices = indices[s]
                classifications = classifications[s]

                # Distribute the sorted spectra to each of `n_pools` in turn to ensure even workload
                data = []
                for pool in range(n_pools):
                    data += [[spectra[i], indices[i], classifications[i]] for i in range(pool, len(spectra), n_pools)]

                print("Processing {} spectra over {} pools".format(n_valid, n_pools))

                with Pool(n_pools) as p:  # Send the job to the `n_pools` pools
                    results = p.map(func, data)  # Map each element of `data` to the first argument of `func`
                    p.close()  # Finished, so no more jobs to come
                    p.join()  # Clean up the closed pools
                    p.clear()  # Remove the server so it can be created again

            else:
                raise TypeError("n_pools must be an integer, got %s" % type(n_pools))

        else:  # Explicit spectrum must be 1D so no loop needed
            results = self._fit(spectra, classification=classifications, spectrum_index=None, **kwargs)

        return results

    def fit_spectrum(self, spectrum, **kwargs):
        """Fits the specified spectrum array.

        Passes the spectrum argument to the :meth:`~mcalf.models.ModelBase.fit` method.
        For easily iterating over a list of spectra.

        Parameters
        ----------
        spectrum : numpy.ndarray, ndim=1
            The explicit spectrum.
        **kwargs : dict, optional
            Extra keyword arguments to pass to :meth:`~mcalf.models.ModelBase.fit`.

        Returns
        -------
        result : :class:`~mcalf.models.FitResult`
            Result of the fit.

        See Also
        --------
        fit : General fitting method.

        Examples
        --------

        Create a basic model:

        >>> import mcalf.models
        >>> import numpy as np
        >>> wavelengths = np.linspace(8541.3, 8542.7, 30)
        >>> model = mcalf.models.ModelBase(original_wavelengths=wavelengths)

        **Quickly provide a spectrum and fit it.** Remember that the model must be optimised for
        the spectra that it is asked to fit. In this example the neural network is not
        called upon to classify the provided spectrum as a classification is provided directly:

        >>> spectrum = np.random.rand(30)
        >>> model.fit_spectrum(spectrum, classifications=0, background=142.2)  # doctest: +SKIP
        Successful FitResult with ________ profile of classification 0

        As the spectrum is provided manually, any background value must also be provided manually.
        Alternatively, the background can be subtracted before passing to the function, as by
        default, no background is subtracted:

        >>> model.fit_spectrum(spectrum - 142.2, classifications=0)  # doctest: +SKIP
        Successful FitResult with ________ profile of classification 0
        """
        return self.fit(spectrum=spectrum, **kwargs)

    def _curve_fit(self, model, spectrum, guess, sigma, bounds, x_scale, time=None, row=None, column=None):
        """:func:`scipy.optimize.curve_fit` wrapper with error handling.

        Passes a certain set of parameters to the :func:`scipy.optimize.curve_fit` function and catches some typical
        errors, presenting a more specific warning message.

        Parameters
        ----------
        model : callable
            The model function, f(x, …). It must take the `ModelBase.constant_wavelenghts` attribute
            as the first argument and the parameters to fit as separate remaining arguments.
        spectrum : array_like
            The dependent data, with length equal to that of the `ModelBase.constant_wavelengths` attribute.
        guess : array_like, optional
            Initial guess for the parameters to fit.
        sigma : array_like
            Determines the uncertainty in the `spectrum`. Used to weight certain regions of the spectrum.
        bounds : 2-tuple of array_like
            Lower and upper bounds on each parameter.
        x_scale : array_like
            Characteristic scale of each parameter.
        time : optional, default=None
            The time index for error handling.
        row : optional, default=None
            The row index for error handling.
        column : optional, default=None
            The column index for error handling.

        Returns
        -------
        fitted_parameters : numpy.ndarray, length=n_parameters
            The parameters that recreate the model fitted to the spectrum.
        success : bool
            Whether the fit was successful or an error had to be handled.

        See Also
        --------
        fit : General fitting method.
        fit_spectrum : Explicit spectrum fitting method.

        Notes
        -----
        More details can be found in the documentation for :func:`scipy.optimize.curve_fit`
        and :func:`scipy.optimize.least_squares`.
        """
        try:  # TODO Investigate if there is a performance gain to setting `check_finite` to False

            fitted_parameters = curve_fit(model, self.constant_wavelengths, spectrum,
                                          p0=guess, sigma=sigma, bounds=bounds, x_scale=x_scale)[0]
            success = True

        except RuntimeError:

            success = False
            location_text = " at ({}, {}, {})".format(time, row, column) if time is not None else ''
            warnings.warn("RuntimeError{}".format(location_text))
            fitted_parameters = np.full_like(guess, np.nan)

        except ValueError as e:

            success = False
            if str(e) != "`x0` violates bound constraints.":
                raise
            else:
                location_text = " at ({}, {}, {})".format(time, row, column) if time is not None else ''
                warnings.warn("`x0` violates bound constraints{}".format(location_text))
                fitted_parameters = np.full_like(guess, np.nan)

        except np.linalg.LinAlgError as e:

            success = False
            if str(e) != "SVD did not converge in Linear Least Squares":
                raise
            else:
                location_text = " at ({}, {}, {})".format(time, row, column) if time is not None else ''
                warnings.warn("SVD did not converge in Linear Least Squares{}.".format(location_text))
                fitted_parameters = np.full_like(guess, np.nan)

        return fitted_parameters, success


# Define each parameter and attribute in an ordered dictionary so definitions can be passed to child objects
DOCS = collections.OrderedDict()
DOCS['original_wavelengths'] = """
    original_wavelengths : array_like
        One-dimensional array of wavelengths that correspond to the uncorrected spectral data."""
DOCS['stationary_line_core'] = """
    stationary_line_core : float, optional, default=None
        Wavelength of the stationary line core."""
DOCS['neural_network'] = """
    neural_network : optional, default=None
        The neural network classifier object that is used to classify spectra. This attribute should be set by a
        child class of :class:`~mcalf.models.ModelBase`."""
DOCS['constant_wavelengths'] = """
    constant_wavelengths : array_like, ndim=1, optional, default= see description
        The desired set of wavelengths that the spectral data should be rescaled to represent. It is assumed
        that these have constant spacing, but that may not be a requirement if you specify your own array.
        The default value is an array from the minimum to the maximum wavelength of `original_wavelengths` in
        constant steps of `delta_lambda`, overshooting the upper bound if the maximum wavelength has not been
        reached."""
DOCS['delta_lambda'] = """
    delta_lambda : float, optional, default=0.05
        The step used between each value of `constant_wavelengths` when its default value has to be calculated."""
DOCS['sigma'] = """
    sigma : optional, default=None
        Sigma values used to weight the fit. This attribute should be set by a child class of
        :class:`~mcalf.models.ModelBase`."""
DOCS['prefilter_response'] = """
    prefilter_response : array_like, length=n_wavelengths, optional, default= see note
        Each constant wavelength scaled spectrum will be corrected by dividing it by this array. If `prefilter_response`
        is not given, and `prefilter_ref_main` and `prefilter_ref_wvscl` are not given, `prefilter_response` will have a
        default value of `None`."""
DOCS['prefilter_ref_main'] = """
    prefilter_ref_main : array_like, optional, default= None
        If `prefilter_response` is not specified, this will be used along with `prefilter_ref_wvscl` to generate the
        default value of `prefilter_response`."""
DOCS['prefilter_ref_wvscl'] = """
    prefilter_ref_wvscl : array_like, optional, default=None
        If `prefilter_response` is not specified, this will be used along with `prefilter_ref_main` to generate the
        default value of `prefilter_response`."""
DOCS['config'] = """
    config : str, optional, default=None
        Filename of a `.yml` file (relative to current directory) containing the initialising parameters for this
        object. Parameters provided explicitly to the object upon initialisation will override any provided in this
        file. All (or some) parameters that this object accepts can be specified in this file, except `neural_network`
        and `config`. Each line of the file should specify a different parameter and be formatted like
        `emission_guess: '[-inf, wl-0.15, 1e-6, 1e-6]'` or `original_wavelengths: 'original.fits'` for example.
        When specifying a string, use 'inf' to represent `np.inf` and 'wl' to represent `stationary_line_core` as shown.
        If the string matches a file, :func:`mcalf.utils.misc.load_parameter()` is used to load the contents
        of the file."""
DOCS['output'] = """
    output : str, optional, default=None
        If the program wants to output data, it will place it relative to the location specified by this parameter.
        Some methods will only save data to a file if this parameter is not `None`. Such cases will be documented
        where relevant."""
DOCS['array'] = """
    array: numpy.ndarray, dimensions are ['time', 'row', 'column', 'spectra']
        Array holding spectra."""
DOCS['background'] = """
    background: numpy.ndarray, dimensions are ['time', 'row', 'column']
        Array holding spectral backgrounds."""

# Form the parameter list
BASE_PARAMETERS = copy.deepcopy(DOCS)
for k in [  # Remove some entries
    'neural_network',
    'array',
    'background',
]:
    del BASE_PARAMETERS[k]
BASE_PARAMETERS_STR = ''.join(BASE_PARAMETERS[i] for i in BASE_PARAMETERS)

# Form the attribute list
BASE_ATTRIBUTES = copy.deepcopy(DOCS)
for k in [  # Remove some entries
    'delta_lambda',
    'prefilter_ref_main',
    'prefilter_ref_wvscl',
    'config',
]:
    del BASE_ATTRIBUTES[k]
BASE_ATTRIBUTES_STR = ''.join(BASE_ATTRIBUTES[i] for i in BASE_ATTRIBUTES)

# Update the docstring with the generated strings
ModelBase.__doc__ = ModelBase.__doc__.replace(
    '${PARAMETERS}',
    BASE_PARAMETERS_STR.lstrip()
)
ModelBase.__doc__ = ModelBase.__doc__.replace(
    '${ATTRIBUTES}',
    BASE_ATTRIBUTES_STR.lstrip()
)

update_signature(ModelBase)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import datetime
# sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../src'))


# -- Project information -----------------------------------------------------

project = 'MCALF'
author = 'Conor D. MacBride & David B. Jess'
copyright = '{}, {}'.format(datetime.datetime.now().year, author)

# The full version, including alpha/beta/rc tags
from pkg_resources import get_distribution
release = get_distribution('mcalf').version


# -- General configuration ---------------------------------------------------

import sphinx_rtd_theme
from sphinx_gallery.sorting import FileNameSortKey

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx_rtd_theme',
    'nbsphinx',
    'sphinx_automodapi.automodapi',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx_gallery.gen_gallery',
]
numpydoc_show_class_members = False
nbsphinx_execute = 'never'
automodsumm_private_methods_of = ['mcalf.models.ModelBase']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

intersphinx_mapping = {
    "python": ('https://docs.python.org/3', None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "matplotlib": ('https://matplotlib.org/', None),
    "numpy": ('https://numpy.org/doc/stable', None),
    "scipy": ('https://docs.scipy.org/doc/scipy/reference', None),
    "sklearn": ('https://scikit-learn.org/stable', None),
}

# sphinx-gallery configuration
sphinx_gallery_conf = {
    # path to your example scripts
    'examples_dirs': ['../examples/gallery'],
    # path to where to save gallery generated output
    'gallery_dirs': ['gallery'],
    # specify that examples should be ordered according to filename
    'within_subsection_order': FileNameSortKey,
    # directory where function/class granular galleries are stored
    'backreferences_dir': 'gen_modules/backreferences',
    # Modules for which function/class level galleries are created.
    'doc_module': ('mcalf',),
    # don't print e.g. sphinx_gallery_thumbnail_number = 2
    'remove_config_comments': True,
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



import copy
import collections

__all__ = ['Parameter', 'ParameterDict', 'OrderedParameterDict',
           'BaseParameterDict', 'SyncedParameters']


class Parameter:
    """A named parameter with a optional value.

    The value can be changed at any time and very basic
    operations can be queued and evaluated on demand.

    Parameters
    ----------
    name : str
        Name of parameter.
    value : int or float, optional, default=None
        Value of parameter. Other types may work but not supported.

    Attributes
    ----------
    name : str
        Name of parameter.
    value : int or float
        Value of parameter, without operations applied.
    operations : list
        List of operations to apply.

    Raises
    ------
    ValueError
        The parameter is evaluated without a value.
    NotImplementedError
        Operations which require parentheses are applied.

    See Also
    --------
    Parameters : Collection of synced ``Parameter`` objects.

    Notes
    -----
    Basic operations can be applied to this object, i.e., ``+``, ``-``, ``*`` and ``/``.
    The ``Parameter`` object must be the left-most entity in an expression.
    Expressions which require parentheses are not supported.
    The only types that are supported in expressions are ``Parameter``, ``float``
    and ``int``.
    Any type should work as a value for ``Parameter`` (such as an array), however,
    only ``float`` and ``int`` are supported.
    See the examples for more details in how it works.

    Examples
    --------
    >>> Parameter('x')
    'x'
    >>> Parameter('x', 12)
    '12'
    >>> Parameter('x') + 1
    'x+1'
    >>> Parameter('x', 1) + 1
    'x+1'
    >>> a = Parameter('y') * 2
    >>> a
    'y*2'
    >>> a.value = 5
    >>> a == 10
    True
    >>> a.eval()
    10
    >>> (a * 10).eval()
    100
    >>> a * 10 + 1.5
    'y*2*10+1.5'
    >>> (a + 10) * 2
    Traceback (most recent call last):
     ...
    NotImplementedError: Parameter equations with brackets are not supported.
    """
    def __init__(self, name, value=None):
        self.name = str(name)
        self.value = value
        self.operations = []

    def __add__(self, other):
        return self.apply_operation('+', other)

    def __sub__(self, other):
        return self.apply_operation('-', other)

    def __mul__(self, other):
        return self.apply_operation('*', other)

    def __truediv__(self, other):
        return self.apply_operation('/', other)

    def apply_operation(self, op, other):
        """Apply an operation to the ``Parameter``.

        Parameters
        ----------
        op : str
            Operation symbol to apply. ``+``, ``-``, ``*`` or ``/``.
        other : float or int
            Number on RHS.

        Returns
        -------
        obj
            A copy of ``self`` with the operation added to the queue.

        Notes
        -----
        You should only use this method directly if a builtin Python operation
        is not otherwise implemented in this class, i.e., do ``a + 1`` and not
        ``a.apply_operation('+', 1)``.
        """
        obj = self.copy()  # Operate on a copy
        obj.operations += [(op, other)]  # Add to queue
        obj.verify()  # Verify no parentheses
        return obj

    def __eq__(self, other):
        # Test equality against the expression's numeric value.
        # If the parameter doesn't have a value,
        # test equality against the string representation
        # of the unevaluated expression.
        try:
            exp = self.eval()
        except ValueError:
            exp = str(self)
        return exp == other

    @property
    def value_or_name(self):
        """Returns its value if set, otherwise returns its name."""
        if self.value is None:
            return self.name
        else:
            return self.value

    def eval(self):
        """Evaluate the expression using the parameter's current value."""
        if self.value is None:
            raise ValueError('Cannot evaluate without a value.')
        return eval(
            str(self),
            {'__builtins__': None},
            {self.name: self.value}
        )

    def __str__(self):
        if len(self.operations) == 0:
            return str(self.value_or_name)
        ret = str(self.name)
        for op, val in self.operations:
            if isinstance(val, Parameter):
                val = val.eval()
            ret += f'{op}{val}'
        return ret

    def __repr__(self):
        return f"'{str(self)}'"

    def verify(self):
        """Enforce order of operations."""
        disable_mul_div = False
        for op, _ in self.operations:
            if op in ('+', '-'):
                disable_mul_div = True
            elif op in ('*', '/') and disable_mul_div:
                raise NotImplementedError(
                    'Parameter equations with brackets are not supported.'
                )

    def copy(self):
        """Return a copy of the parameter."""
        return self.__copy__()

    def __copy__(self):
        obj = self.__class__(copy.copy(self.name), copy.copy(self.value))
        obj.operations = copy.deepcopy(self.operations)
        return obj

    def __deepcopy__(self, memodict={}):
        return self.__copy__()


class SyncedParameters:
    """Database for keeping :class:`Parameter` objects in sync.

    :class:`Parameter` objects with the same name can be linked with this class
    and kept in sync with each other.

    Examples
    --------
    >>> a = Parameter('x') + 1
    >>> b = Parameter('x', 2) * 3
    >>> c = Parameter('y')

    Now that :class:`Parameter` objects have been created,
    we can keep them in sync.

    >>> s = SyncedParameters()
    >>> s.track_object(a)
    >>> s.track_object(b)
    >>> s.track_object(c)

    Notice how the value of `a` has been synced to match the
    value provided in `b`.

    >>> a == 3
    True

    The value of a named parameter can be updated and the
    change is propagated to all :class:`Parameter` objects
    of that name.

    >>> s.update_parameter('x', 3)
    >>> a == 4
    True
    >>> b == 9
    True
    """
    @property
    def _tracked(self):
        """Dictionary of tracked :class:`Parameters` grouped by name.

        Notes
        -----
        Has the form: ``{Parameter.name: {'value': Parameter.value, 'objects': List[Parameter]}}``.
        All ``Parameter.value`` of the same ``Parameter.name`` are
        identical in both ``'value'`` and all in ``'objects'``.
        """
        if not hasattr(self, '_tracked_objs'):
            self._tracked_objs = {}
        return self._tracked_objs

    def track_object(self, obj):
        """Add a new :class:`Parameter` object to keep in sync.

        Parameters
        ----------
        obj : :class:`Parameter`
            The parameter object to keep in sync.

        Notes
        -----
        Any objects added must not have a value that contradicts the existing
        value of the parameter (if set). Objects added which have no value set
        will inherit the value of the other parameters of the same name.
        """
        if self.exists(obj.name):  # If parameter already registered...
            if obj.value is not None:  # ...copy incoming value to existing.
                # (No more than one value should be associated with each name.)
                if self.has_value(obj.name) and obj.value is not self.get_parameter(obj.name):
                    raise ValueError(f"'{obj.name}' value {obj.value} does not match "
                                     f"existing value {self.get_parameter(obj.name)}.")
                self.update_parameter(obj.name, obj.value)
            elif self.has_value(obj.name):  # ...copy existing value to incoming.
                obj.value = self.get_parameter(obj.name)
            self._tracked[obj.name]['objects'] += [obj]
        else:
            self._tracked[obj.name] = {'value': obj.value, 'objects': [obj]}

    def update_parameter(self, name, value):
        """Update the value of a named parameter.

        All tracked :class:`Parameter` objects of this name will be updated.

        Parameters
        ----------
        name : string
            Name of parameter to update.
        value : float or int
            New value of parameter. Other types may work.
        """
        self._tracked[name]['value'] = value
        for obj in self._tracked[name]['objects']:
            obj.value = value

    def get_parameter(self, name):
        """Get the value of the named parameter.

        Parameters
        ----------
        name : str
            Name of parameter to get value of.

        Returns
        -------
        value
            The current value of the parameter. ``None`` if not set.
        """
        return self._tracked[name]['value']

    def has_value(self, name: str) -> bool:
        """Whether the named parameter has a value set."""
        return False if self.get_parameter(name) is None else True

    def exists(self, name: str) -> bool:
        """Whether any parameters of a particular name are tracked."""
        try:
            self._tracked[name]
        except KeyError:
            return False
        else:
            return True


class BaseParameterDict(SyncedParameters):
    """
    A base class for dictionaries of :class:`Parameter` objects.

    The same parameters existing across multiple dictionary
    values can be kept in sync with each other.
    For a :class:`Parameter` object to be kept in sync it must be
    located in the dictionary such that ``{key: Parameter}`` or
    ``{key: List[Parameter]}``.
    """
    def __setitem__(self, key, value):
        if isinstance(value, Parameter):
            self.track_object(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, Parameter):
                    self.track_object(item)
        super().__setitem__(key, value)

    def eval(self):
        """Return a copy of the dictionary with all :class:`Parameter` objects evaluated."""
        edict = self.__class__.__bases__[-1]()
        for key, value in self.items():
            if isinstance(value, Parameter):
                value = value.eval()
            elif isinstance(value, list):
                lst = []
                for i, v in enumerate(value):
                    if isinstance(v, Parameter):
                        lst += [v.eval()]
                    else:
                        lst += [v]
                value = lst
            edict[key] = value
        return edict


class ParameterDict(BaseParameterDict, collections.UserDict):
    """
    An unordered dictionary of :class:`Parameter` objects.

    The same parameters existing across multiple dictionary
    values can be kept in sync with each other.
    For a :class:`Parameter` object to be kept in sync it must be
    located in the dictionary such that ``{key: Parameter}`` or
    ``{key: List[Parameter]}``.

    Examples
    --------
    >>> d = ParameterDict({
    ...     'a': Parameter('x') + 1,
    ...     'b': [2, Parameter('x'), 5],
    ...     'c': {1, 2, 3},
    ... })
    >>> d
    {'a': 'x+1', 'b': [2, 'x', 5], 'c': {1, 2, 3}}
    >>> d.update_parameter('x', 1)
    >>> d.eval()
    {'a': 2, 'b': [2, 1, 5], 'c': {1, 2, 3}}
    """


class OrderedParameterDict(BaseParameterDict, collections.OrderedDict):
    """
    An ordered dictionary of :class:`Parameter` objects.

    The same parameters existing across multiple dictionary
    values can be kept in sync with each other.
    For a :class:`Parameter` object to be kept in sync it must be
    located in the dictionary such that ``{key: Parameter}`` or
    ``{key: List[Parameter]}``.

    Examples
    --------
    >>> d = OrderedParameterDict([
    ...     ('a', Parameter('x') + 1),
    ...     ('b', [2, Parameter('x'), 5]),
    ...     ('c', {1, 2, 3}),
    ... ])
    >>> d
    OrderedParameterDict([('a', 'x+1'), ('b', [2, 'x', 5]), ('c', {1, 2, 3})])
    >>> d.update_parameter('x', 1)
    >>> d.eval()
    OrderedDict([('a', 2), ('b', [2, 1, 5]), ('c', {1, 2, 3})])
    """



import numpy as np
import pytest

from mcalf.utils.mask import genmask, radial_distances


def test_radial_distances():
    res = radial_distances(4, 9)
    truth = np.array([[4.27200187, 4.03112887, 4.03112887, 4.27200187],
                      [3.35410197, 3.04138127, 3.04138127, 3.35410197],
                      [2.5, 2.06155281, 2.06155281, 2.5],
                      [1.80277564, 1.11803399, 1.11803399, 1.80277564],
                      [1.5, 0.5, 0.5, 1.5],
                      [1.80277564, 1.11803399, 1.11803399, 1.80277564],
                      [2.5, 2.06155281, 2.06155281, 2.5],
                      [3.35410197, 3.04138127, 3.04138127, 3.35410197],
                      [4.27200187, 4.03112887, 4.03112887, 4.27200187]])
    assert res == pytest.approx(truth)


def test_genmask():
    res = genmask(4, 9, radius=2.2)
    truth = np.array([[False, False, False, False],
                      [False, False, False, False],
                      [False, True, True, False],
                      [True, True, True, True],
                      [True, True, True, True],
                      [True, True, True, True],
                      [False, True, True, False],
                      [False, False, False, False],
                      [False, False, False, False]])
    assert np.sum(res == truth) == 4 * 9

    res = genmask(4, 9, radius=2.2, right_shift=1, up_shift=-2)
    truth = np.array([[False, False, True, True],
                      [True, True, True, True],
                      [True, True, True, True],
                      [True, True, True, True],
                      [False, False, True, True],
                      [False, False, False, False],
                      [False, False, False, False],
                      [False, False, False, False],
                      [False, False, False, False]])
    assert np.sum(res == truth) == 4 * 9



import warnings

import numpy as np
from scipy.integrate import IntegrationWarning, quad, quad_vec

# Load the C library
import ctypes
import os.path
from pathlib import Path
# # Commands to manually generate
# gcc -Wall -fPIC -c voigt.c
# gcc -shared -o libvoigt.so voigt.o
dllabspath = Path(os.path.dirname(os.path.abspath(__file__)))  # Path to libraries directory
try:
    libfile = [str(i) for i in dllabspath.glob('ext_voigtlib.*.so')][0]  # Select first (and only) library
    lib = ctypes.CDLL(libfile)  # Load the library
    lib.func.restype = ctypes.c_double  # Specify the expected result type
    lib.func.argtypes = (ctypes.c_int, ctypes.c_double)  # Specify the type of the input parameters
    cvoigt = lib.func  # Create alias for the specific function used in functions below
except IndexError:  # File does not exist
    warnings.warn("Could not locate the external C library. Further use of `clib` will fail!")

###
# readthedocs.org does not support clib (force clib=False)
import os  # noqa: E402
not_on_rtd = os.environ.get('READTHEDOCS') != 'True'
rtd = {}
if not not_on_rtd:  # Reduce computation time (and accuracy) of no clib version
    rtd = {'epsabs': 1.49e-1, 'epsrel': 1.49e-4}
###

# Parameters for `voigt_approx_nobg` and other approx. Voigt functions
params = np.array([[-1.2150, -1.3509, -1.2150, -1.3509],
                   [1.2359, 0.3786, -1.2359, -0.3786],
                   [-0.3085, 0.5906, -0.3085, 0.5906],
                   [0.0210, -1.1858, -0.0210, 1.1858]])
sqrt_ln2 = np.sqrt(np.log(2))
sqrt_pi = np.sqrt(np.pi)
A, B, C, D = params


__all__ = ['voigt_approx_nobg', 'voigt_approx', 'double_voigt_approx_nobg', 'double_voigt_approx',
           'voigt_nobg', 'voigt', 'double_voigt_nobg', 'double_voigt']


def voigt_approx_nobg(x, a, b, s, g):
    """Voigt function (efficient approximation) with no background (Base approx. Voigt function).

    This is the base for all other approximated Voigt functions. Not implemented in any models yet as initial tests
    exhibited slow convergence.

    Parameters
    ----------
    ${SINGLE_VOIGT}

    ${EXTRA_APPROX}
    """
    fwhm_g = 2 * s * np.sqrt(2 * np.log(2))
    fwhm_l = 2 * g
    xx = (x - b) * 2 * sqrt_ln2 / fwhm_g
    xx = xx[..., np.newaxis]
    yy = fwhm_l * sqrt_ln2 / fwhm_g
    yy = yy[..., np.newaxis]
    v = np.sum((C * (yy - A) + D * (xx - B)) / ((yy - A) ** 2 + (xx - B) ** 2), axis=-1)
    return fwhm_l * a * sqrt_pi / fwhm_g * v


def voigt_approx(x, a, b, s, g, d):
    """Voigt function (efficient approximation) with background.

    Parameters
    ----------
    ${SINGLE_VOIGT}
    ${BACKGROUND}

    ${EXTRA_APPROX}
    """
    return voigt_approx_nobg(x, a, b, s, g) + d


def double_voigt_approx_nobg(x, a1, b1, s1, g1, a2, b2, s2, g2):
    """Double Voigt function (efficient approximation) with no background.

    Parameters
    ----------
    ${DOUBLE_VOIGT}

    ${EXTRA_APPROX}
    """
    return voigt_approx_nobg(x, a1, b1, s1, g1) + voigt_approx_nobg(x, a2, b2, s2, g2)


def double_voigt_approx(x, a1, b1, s1, g1, a2, b2, s2, g2, d):
    """Double Voigt function (efficient approximation) with background.

    Parameters
    ----------
    ${DOUBLE_VOIGT}
    ${BACKGROUND}

    ${EXTRA_APPROX}
    """
    return voigt_approx_nobg(x, a1, b1, s1, g1) + voigt_approx_nobg(x, a2, b2, s2, g2) + d


def voigt_nobg(x, a, b, s, g, clib=True):
    """Voigt function with no background (Base Voigt function).

    This is the base of all the other Voigt functions.

    Parameters
    ----------
    ${SINGLE_VOIGT}
    ${CLIB}

    ${EXTRA_STD}
    """
    warnings.filterwarnings("ignore", category=IntegrationWarning)
    u = x - b
    if clib and not_on_rtd:
        i = [quad(cvoigt, -np.inf, np.inf, args=(v, s, g), epsabs=1.49e-1, epsrel=1.49e-4)[0] for v in u]
    else:
        i = quad_vec(lambda y: np.exp(-y**2 / (2 * s**2)) / (g**2 + (u - y)**2), -np.inf, np.inf, **rtd)[0]
    const = g / (s * np.sqrt(2 * np.pi**3))
    return a * const * np.array(i)


def voigt(x, a, b, s, g, d, clib=True):
    """Voigt function with background.

    Parameters
    ----------
    ${SINGLE_VOIGT}
    ${BACKGROUND}
    ${CLIB}

    ${EXTRA_STD}
    """
    return voigt_nobg(x, a, b, s, g, clib) + d


def double_voigt_nobg(x, a1, b1, s1, g1, a2, b2, s2, g2, clib=True):
    """Double Voigt function with no background.

    Parameters
    ----------
    ${DOUBLE_VOIGT}
    ${CLIB}

    ${EXTRA_STD}
    """
    return voigt_nobg(x, a1, b1, s1, g1, clib) + voigt_nobg(x, a2, b2, s2, g2, clib)


def double_voigt(x, a1, b1, s1, g1, a2, b2, s2, g2, d, clib=True):
    """Double Voigt function with background.

    Parameters
    ----------
    ${DOUBLE_VOIGT}
    ${BACKGROUND}
    ${CLIB}

    ${EXTRA_STD}
    """
    return double_voigt_nobg(x, a1, b1, s1, g1, a2, b2, s2, g2, clib) + d


# Define "Parameters" options
__input_x = """
    x : numpy.ndarray
        Wavelengths to evaluate Voigt function at."""
__single_voigt = __input_x + """
    a : float
        Amplitude of the Lorentzian.
    b : float
        Central line core.
    s : float
        Sigma (for Gaussian).
    g : float
        Gamma (for Lorentzian)."""
__double_voigt = __input_x + """
    a1 : float
        Amplitude of 1st Voigt function.
    b1 : float
        Central line core of 1st Voigt function.
    s1 : float
        Sigma (for Gaussian) of 1st Voigt function.
    g1 : float
        Gamma (for Lorentzian) of 1st Voigt function.
    a2 : float
        Amplitude of 2nd Voigt function.
    b2 : float
        Central line core of 2nd Voigt function.
    s2 : float
        Sigma (for Gaussian) of 2nd Voigt function.
    g2 : float
        Gamma (for Lorentzian) of 2nd Voigt function."""
__background = """
    d : float
        Background."""
__clib = """
    clib : bool, optional, default=True
        Whether to use the complied C library or a slower Python version. If using the C library, the accuracy
        of the integration is reduced to give the code a significant speed boost. Python version can be used when
        speed is not a priority. Python version will remove deviations that are sometimes present around the wings
        due to the reduced accuracy."""

# Define "Returns"
__returns = """
    Returns
    -------
    result : numpy.ndarray, shape=`x.shape`
        The value of the Voigt function here.

    """

# Define "Notes" (and "References") section
__notes_approx = """
    Notes
    -----
    This algorithm is taken from A. B. McLean et al. [1]_.

    References
    ----------
    .. [1] A. B. McLean, C. E. J. Mitchell and D. M. Swanston, "Implementation of an efficient analytical
      approximation to the Voigt function for photoemission lineshape analysis," Journal of Electron Spectroscopy and
      Related Phenomena, vol. 69, pp. 125-132, 1994. https://doi.org/10.1016/0368-2048(94)02189-7"""
__notes_std = """
    Notes
    -----
    More information on the Voigt function can be found here: https://en.wikipedia.org/wiki/Voigt_profile"""


# Define special "See Also" options
__see_also_approx = [
    '    voigt_approx_nobg : Base approximated Voigt function with no background.',
    '    voigt_approx : Approximated Voigt function with background added.',
    '    double_voigt_approx_nobg : Two approximated Voigt functions added together.',
    '    double_voigt_approx : Two approximated Voigt functions and a background added together.',
]
__see_also_std = [
    '    voigt_nobg : Base Voigt function with no background.',
    '    voigt : Voigt function with background added.',
    '    double_voigt_nobg : Two Voigt functions added together.',
    '    double_voigt : Two Voigt function and a background added together.',
]
# Extract the function name for easy exclusion of item
__see_also_approx = [(i.split(':')[0].strip(), i) for i in __see_also_approx]
__see_also_std = [(i.split(':')[0].strip(), i) for i in __see_also_std]
__see_also = __see_also_approx + __see_also_std


def __rm_self(func, items):
    """Return the "See Also" section with the current function removed."""
    ret = [i[1] for i in items if i[0] != func]
    return 'See Also\n    --------\n    ' + '\n'.join(ret).lstrip() + '\n'


def __extra_approx(func):
    """Merge common approx functions sections."""
    return __returns + __rm_self(func.__name__, __see_also) + __notes_approx  # noqa: F821


def __extra_std(func):
    """Merge common standard functions sections."""
    return __returns + __rm_self(func.__name__, __see_also_std) + __notes_std  # noqa: F821


for f in [voigt_approx_nobg, voigt_approx, double_voigt_approx_nobg, double_voigt_approx,
          voigt_nobg, voigt, double_voigt_nobg, double_voigt]:
    f.__doc__ = f.__doc__.replace('${SINGLE_VOIGT}', __single_voigt.lstrip())
    f.__doc__ = f.__doc__.replace('${DOUBLE_VOIGT}', __double_voigt.lstrip())
    f.__doc__ = f.__doc__.replace('${BACKGROUND}', __background.lstrip())
    f.__doc__ = f.__doc__.replace('${CLIB}', __clib.lstrip())
    f.__doc__ = f.__doc__.replace('${EXTRA_APPROX}', __extra_approx(f).lstrip())
    f.__doc__ = f.__doc__.replace('${EXTRA_STD}', __extra_std(f).lstrip())

del __input_x
del __single_voigt
del __double_voigt
del __background
del __clib
del __returns
del __notes_approx
del __notes_std
del __see_also_approx
del __see_also_std
del __see_also
del __rm_self
del __extra_approx
del __extra_std



import numpy as np
import pytest

from mcalf.models import FitResult, FitResults
from mcalf.models import ModelBase as DummyModel

fitted_parameters = [1, 2, 1000.2, 1001.8, 5]
fit_info = {'chi2': 1.4, 'classification': 2, 'profile': 'abc',
            'success': True, 'index': [123, 456, 789]}


def test_fitresult_passthrough():
    fit = FitResult(fitted_parameters, fit_info)
    assert fit.parameters == [1, 2, 1000.2, 1001.8, 5]
    assert len(fit) == 5
    assert fit.chi2 == 1.4
    assert fit.classification == 2
    assert fit.profile == 'abc'
    assert isinstance(fit.success, bool) and fit.success
    assert fit.index == [123, 456, 789]

    # Test that the string representation can be formed without error
    repr(fit)
    fit.index = [None]*3
    repr(fit)


def test_fitresult_velocity():
    m = DummyModel(original_wavelengths=[1000.4, 1000.6])
    m.stationary_line_core = 1000.5
    m.quiescent_wavelength = 2
    m.active_wavelength = 3
    fit = FitResult(fitted_parameters, fit_info)
    assert fit.velocity(m, vtype='quiescent') == pytest.approx(-89.95502249)
    assert fit.velocity(m, vtype='active') == pytest.approx(389.80509745)

    # Ensure nan is returned if no active component fitted
    fitted_parameters_trim = fitted_parameters[:3]
    fit = FitResult(fitted_parameters_trim, fit_info)
    vel = fit.velocity(m, vtype='active')
    assert vel != vel  # assert is nan

    # Ensure an invalid velocity type is detected
    with pytest.raises(ValueError):
        vel = fit.velocity(m, vtype='unknown-vtype')


def test_fitresults_init():
    fits = FitResults((49, 52), 4, time=12)
    assert fits.parameters.shape == (49, 52, 4)
    assert fits.chi2.shape == (49, 52)
    assert fits.classifications.shape == (49, 52)
    assert fits.profile.shape == (49, 52)
    assert fits.success.shape == (49, 52)
    assert fits.time == 12

    with pytest.raises(TypeError):  # Should be a tuple
        fits = FitResults(10, 3)

    with pytest.raises(TypeError):  # Should be a tuple of length 2
        fits = FitResults((10, 32, 53), 8)

    with pytest.raises(ValueError):  # Should be an integer >= 1
        fits = FitResults((10, 5), 5.5)
    with pytest.raises(ValueError):  # Should be an integer >= 1
        fits = FitResults((10, 5), 0)


def test_fitresults_append():

    # Create dummy fit results
    fit1 = FitResult(
        [2, 6, 254.6, 963.4],
        {'chi2': 7.43, 'classification': 4, 'profile': 'absorption',
            'success': True, 'index': [12, 34, 81]}
    )
    fit2 = FitResult(
        [9, 2, 724.32, 134.8],
        {'chi2': 1.34, 'classification': 2, 'profile': 'emission',
            'success': True, 'index': [12, 0, 99]}
    )
    fit3 = FitResult(
        [1, 8, 932.1, 327.5, 3.7, 9, 2, 0.2],
        {'chi2': 0.79, 'classification': 1, 'profile': 'both',
            'success': False, 'index': [12, 99, 0]}
    )
    fit4 = FitResult(  # With incorrect time index
        [6, 4, 356.2, 738.5],
        {'chi2': 8.2, 'classification': 3, 'profile': 'absorption',
            'success': True, 'index': [3, 0, 25]}
    )
    fit5 = FitResult(  # With unknown profile name
        [5, 3, 256.2, 628.5],
        {'chi2': 8.1, 'classification': 3, 'profile': 'continuum',
            'success': True, 'index': [12, 10, 40]}
    )

    # Initialise empty FitResults object
    fits = FitResults((100, 100), 8, time=12)

    # Append dummy fits
    fits.append(fit1)
    fits.append(fit2)
    fits.append(fit3)

    with pytest.raises(ValueError):  # Time index does not match
        fits.append(fit4)

    with pytest.raises(ValueError):  # Unknown profile
        fits.append(fit5)

    assert all([a == b for a, b in zip(fits.parameters[34, 81][:4], fit1.parameters)])
    assert fits.chi2[34, 81] == fit1.chi2
    assert fits.classifications[34, 81] == fit1.classification
    assert fits.profile[34, 81] == fit1.profile
    assert fits.success[34, 81] == fit1.success

    assert all([a == b for a, b in zip(fits.parameters[0, 99][4:], fit2.parameters)])
    assert fits.chi2[0, 99] == fit2.chi2
    assert fits.classifications[0, 99] == fit2.classification
    assert fits.profile[0, 99] == fit2.profile
    assert fits.success[0, 99] == fit2.success

    assert all([a == b for a, b in zip(fits.parameters[99, 0], fit3.parameters)])
    assert fits.chi2[99, 0] == fit3.chi2
    assert fits.classifications[99, 0] == fit3.classification
    assert fits.profile[99, 0] == fit3.profile
    assert fits.success[99, 0] == fit3.success


def test_fitresults_velocities():
    m = DummyModel(original_wavelengths=[1000.4, 1000.6])
    m.stationary_line_core = 1000.5
    m.quiescent_wavelength = 0
    m.active_wavelength = 1
    fits = FitResults((4, 4), 2)
    fits.parameters = np.array([
        [[1000.2, 192.4], [826.5, 534.23], [8365.86, 1252.32], [1532.3, 2152.3]],
        [[978.73, 753.52], [1253.5, 1329.3], [6423.4, 2355.45], [12.53, 2523.3]],
        [[825.8, 862.5], [1759.5, 1000.9], [2633.4, 234.43], [2535.353, 152.34]],
        [[896.53, 153.2], [1224.3, 1111.11], [634.54, 2353.97], [242.35, 763.4]]
    ])

    truth_quiescent = np.array([[-8.99550225e+01, -5.21739130e+04, 2.20850375e+06, 1.59460270e+05],
                                [-6.52773613e+03, 7.58620690e+04, 1.62605697e+06, -2.96242879e+05],
                                [-5.23838081e+04, 2.27586207e+05, 4.89625187e+05, 4.60225787e+05],
                                [-3.11754123e+04, 6.71064468e+04, -1.09733133e+05, -2.27331334e+05]])

    truth_active = np.array([[-2.42308846e+05, -1.39811094e+05, 7.55082459e+04, 3.45367316e+05],
                             [-7.40569715e+04, 9.85907046e+04, 4.06281859e+05, 4.56611694e+05],
                             [-4.13793103e+04, 1.19940030e+02, -2.29706147e+05, -2.54320840e+05],
                             [-2.54062969e+05, 3.31664168e+04, 4.05838081e+05, -7.10944528e+04]])

    result_quiescent = fits.velocities(m, vtype='quiescent')
    result_active = fits.velocities(m, vtype='active')

    with pytest.raises(ValueError):
        fits.velocities(m, vtype='unknown-vtype')

    assert result_quiescent == pytest.approx(truth_quiescent)
    assert result_active == pytest.approx(truth_active)



import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pytest

from mcalf.visualisation import bar, init_class_data, plot_class_map, plot_classifications

from ..helpers import class_map, figure_test


def spectra(n=20, w=10, low=0, high=6):
    """Generate an array y of n spectra each with w wavelength points
    along the second dimension. Array r of length n has random integers
    from [low, high)."""
    np.random.seed(4321)
    a, b, s = np.random.rand(3, n, 1)

    a = a - 0.5

    width_deviation = w / 5
    b = width_deviation * b - (width_deviation / 2) + w // 2

    s = 2 * s + 2

    x = np.array([np.arange(w)])

    y = a * np.exp(- (x - b) ** 2 / (2 * s ** 2)) + 0.5

    r = np.random.randint(low, high, len(y))

    return y, r


def test_plot_classifications():

    with pytest.raises(TypeError) as e:
        plot_classifications([[0, 1], [1, 2]], np.array([3, 4]))
    assert '`spectra` must be a numpy.ndarray' in str(e.value)

    with pytest.raises(TypeError) as e:
        plot_classifications(np.array([[0, 1], [1, 2]]), [3, 4])
    assert '`labels` must be a numpy.ndarray' in str(e.value)

    with pytest.raises(TypeError) as e:
        plot_classifications(np.array([0, 1]), np.array([3, 4]))
    assert '`spectra` must be a 2D array' in str(e.value)

    for r in (np.array([[3, 4]]), np.array([3.5, 4])):
        with pytest.raises(TypeError) as e:
            plot_classifications(np.array([[0, 1], [1, 2]]), r)
        assert '`labels` must be a 1D array of integers' in str(e.value)

    with pytest.raises(ValueError) as e:
        plot_classifications(np.array([[0, 1], [1, 2]]), np.array([3, 4, 6]))
    assert '`spectra` and `labels` must be the same length' in str(e.value)

    s, r = spectra(10, 5, 0, 6)

    with pytest.raises(ValueError) as e:
        plot_classifications(s, r, nrows=3, ncols=2)
    assert 'Both `nrows` and `ncols` cannot be given together' in str(e.value)

    with pytest.raises(TypeError) as e:
        plot_classifications(s, r, nrows=2.5)
    assert '`nrows` must be an integer' in str(e.value)

    with pytest.raises(TypeError) as e:
        plot_classifications(s, r, ncols=3.5)
    assert '`ncols` must be an integer' in str(e.value)

    for nlines in (3.5, -3, 0):
        with pytest.raises(TypeError) as e:
            plot_classifications(s, r, nlines=nlines)
        assert '`nlines` must be a positive integer' in str(e.value)

    s, r = spectra(100, 5, 2, 7)  # 5 plots

    with pytest.raises(ValueError) as e:
        plot_classifications(s, r, nrows=6)  # 6x1, not 5x1
    assert '`nrows` is larger than it needs to be' in str(e.value)

    with pytest.raises(ValueError) as e:
        plot_classifications(s, r, ncols=4)  # 2x4, not 2x3
    assert '`ncols` is larger than it needs to be' in str(e.value)


@figure_test
def test_plot_classifications_5plots(pytestconfig):
    s, r = spectra(40, 30, 0, 5)  # 5 plots
    plot_classifications(s, r)


@figure_test
def test_plot_classifications_4plots(pytestconfig):
    s, r = spectra(40, 30, 1, 5)  # 4 plots
    plot_classifications(s, r)


@figure_test
def test_plot_classifications_3plots(pytestconfig):
    s, r = spectra(40, 30, 2, 5)  # 3 plots
    plot_classifications(s, r, show_labels=False)


@figure_test
def test_plot_classifications_1plot(pytestconfig):
    s, r = spectra(40, 30, 3, 4)  # 1 plot
    plot_classifications(s, r)


@figure_test
def test_plot_classifications_3x2_5(pytestconfig):
    s, r = spectra(40, 30, 0, 5)  # 5 plots
    plot_classifications(s, r, nrows=3)


@figure_test
def test_plot_classifications_4x1(pytestconfig):
    s, r = spectra(40, 30, 1, 5)  # 4 plots
    plot_classifications(s, r, ncols=1)


@figure_test
def test_plot_classifications_3x2(pytestconfig):
    s, r = spectra(40, 30, 2, 8)  # 6 plots
    plot_classifications(s, r, nrows=3)


@figure_test
def test_plot_classifications_not01(pytestconfig):
    s, r = spectra(40, 30, 3, 5)  # 2 plots
    s[r == 4] += 0.5  # classification 4 not in [0, 1]
    plot_classifications(s, r)


def test_no_args():
    # Custom errors as not required if data keyword argument given.

    with pytest.raises(TypeError) as e:
        bar()
    assert "bar() missing 1 required positional argument: 'class_map'" == str(e.value)

    with pytest.raises(TypeError) as e:
        plot_class_map()
    assert "plot_class_map() missing 1 required positional argument: 'class_map'" == str(e.value)


@figure_test
def test_bar(pytestconfig):
    m = class_map(30, 10, 20, 8)
    bar(m)


@figure_test
def test_bar_range(pytestconfig):
    m = class_map(30, 10, 20, 8)
    bar(m, vmin=2, vmax=4)


@figure_test
def test_bar_reduce(pytestconfig):
    m = class_map(30, 10, 20, 8)
    bar(m, reduce=False)


@figure_test
def test_bar_allbad(pytestconfig):
    m = class_map(30, 10, 20, 8)
    m[:] = -1
    bar(m)


@figure_test
def test_plot_class_map(pytestconfig):
    m = class_map(30, 10, 20, 8)
    plot_class_map(m)


@figure_test
def test_plot_class_map_nocolorbar(pytestconfig):
    m = class_map(30, 10, 20, 8)
    plot_class_map(m, show_colorbar=False)


@figure_test
def test_plot_class_map_range(pytestconfig):
    m = class_map(30, 10, 20, 8)
    plot_class_map(m, vmin=2, vmax=4)


@figure_test
def test_plot_class_map_units(pytestconfig):
    m = class_map(30, 10, 20, 8)
    plot_class_map(m, resolution=(2.5 * u.m, 3.5 * u.km), offset=(5, 10),
                   dimension=('dim x', 'dim y'))


@figure_test
def test_plot_class_map_allbad(pytestconfig):
    m = class_map(30, 10, 20, 8)
    m[:] = -1
    plot_class_map(m)


@figure_test
def test_plot_bar_and_class_map(pytestconfig):
    m = class_map(30, 10, 20, 8)
    fig, ax = plt.subplots(1, 2)
    data = init_class_data(m, ax=ax[0], colorbar_settings={'ax': ax, 'location': 'bottom'})
    plot_class_map(data=data, ax=ax[0])
    bar(data=data, ax=ax[1])



import inspect

import numpy as np
import pytest
from astropy.io import fits

from mcalf.utils.collections import OrderedParameterDict, Parameter
from mcalf.utils.misc import load_parameter, make_iter, merge_results, update_signature

from ..helpers import data_path_function

data_path = data_path_function('utils')


def test_make_iter():
    i = [(1, 2, 3), (4,), [9, "a"], [23], "abc"]
    assert make_iter(*i) == i
    assert make_iter(1, [2], 3) == [[1], [2], [3]]


def test_load_parameter_string():
    # Need to test loading files
    i = "  [  42 ,  9.2, 6 +wl - 2.3, 12.5 + wl,wl, inf, -inf, inf+   4., 5.2-inf] "
    inf = float('inf')
    truth = [42.0, 9.2, 6.9, 15.7, 3.2, inf, -inf, inf, -inf]
    assert load_parameter(i, wl=3.2) == pytest.approx(truth)

    # Everything should be converted to floats
    assert isinstance(load_parameter("wl", wl=int(5)), float)


def test_load_parameter_file():

    truth = np.array([243.23, 62, 2523.43, 0, -62.2, np.inf, np.nan], dtype=np.float64)

    # Unsupported extension given
    with pytest.raises(ValueError):
        load_parameter(data_path('test_load_parameter_file.unsupported'))

    # Test each file type
    for ext in ['npy', 'csv', 'fits', 'sav']:
        res = load_parameter(data_path(f'test_load_parameter_file.{ext}'))
        assert res[:-1] == pytest.approx(truth[:-1])
        assert np.isnan(res[-1])

    # Test CSV with defined wl
    res = load_parameter(data_path('test_load_parameter_file_wl.csv'), wl=2523.43)
    assert res[:-1] == pytest.approx(truth[:-1])
    assert np.isnan(res[-1])

    # Test CSV with expected exceptions
    for e in ['typeerror', 'syntaxerror']:
        with pytest.raises(SyntaxError):  # Note: the TypeError is converted into a SyntaxError
            load_parameter(data_path(f'test_load_parameter_file_{e}.csv'), wl=2523.43)


def test_merge_results(tmp_path):

    # Compatible files to test merging
    compatible_files = [
        data_path('test_merge_results_1.fits'),
        data_path('test_merge_results_2.fits'),
        data_path('test_merge_results_3.fits'),
    ]

    # Merge and save
    output_file = tmp_path / "test_merge_results_output.fits"
    merge_results(compatible_files, output_file)

    # Compare merged files to expected merge
    test = fits.open(output_file, mode='readonly')
    verify = fits.open(data_path('test_merge_results_all.fits'), mode='readonly')
    # Diff ignoring checksums as too strict (compare values instead)
    diff = fits.FITSDiff(test, verify, ignore_keywords=['CHECKSUM', 'DATASUM'])
    assert diff.identical  # If this fails tolerances *may* need to be adjusted

    # Incompatible files to test merging
    incompatible_files = [
        data_path('test_merge_results_1.fits'),
        data_path('test_merge_results_2.fits'),
        data_path('test_merge_results_3.fits'),
        data_path('test_merge_results_2.fits'),  # Duplicate (overlapping) file should fail
    ]

    # Merge (should fail before saving)
    with pytest.raises(ValueError):
        merge_results(incompatible_files, output_file)

    # Compatible files but wrong time for one (should fail)
    compatible_files_wrongtime = [
        data_path('test_merge_results_1.fits'),
        data_path('test_merge_results_2_wrongtime.fits'),
        data_path('test_merge_results_3.fits'),
    ]

    # Merge (should fail before saving)
    with pytest.raises(ValueError):
        merge_results(compatible_files_wrongtime, output_file)

    # Must provide a list of multiple files
    with pytest.raises(TypeError):  # single string
        merge_results(data_path('test_merge_results_1.fits'), output_file)
    with pytest.raises(TypeError):  # list of length 1
        merge_results([data_path('test_merge_results_1.fits')], output_file)

    # The extra HDU should cause an error
    with pytest.raises(ValueError) as excinfo:
        merge_results([
            data_path('test_merge_results_2.fits'),
            data_path('test_merge_results_1_extrahdu.fits'),
        ], output_file)
    assert 'nexpected' in str(excinfo.value)  # "Unexpected"
    # reverse (now extra is in first file)
    with pytest.raises(ValueError) as excinfo:
        merge_results([
            data_path('test_merge_results_1_extrahdu.fits'),
            data_path('test_merge_results_2.fits'),
        ], output_file)
    assert 'nexpected' in str(excinfo.value)  # "Unexpected"


def test_update_signature():

    class A:
        default_kwargs = OrderedParameterDict([
            ('c', Parameter('hello', 15)),
            ('d', [1, Parameter('world'), 3, Parameter('hello')]),
        ])

        def __init__(self, a=None, **kwargs):
            pass

    class B(A):
        default_kwargs = OrderedParameterDict([
            ('e', Parameter('hello') + 12),
            ('d', [Parameter('world') * 3, 3, Parameter('hello'), 6.2]),
        ])

        def __init__(self, b=1, **kwargs):
            pass

    update_signature(A)
    assert str(inspect.signature(A.__init__)) == \
           "(*, c='15', d=[1, 'world', 3, '15'], a=None)"

    update_signature(B)
    assert str(inspect.signature(B.__init__)) == \
           "(*, c='15', a=None, e='hello+12', d=['world*3', 3, 'hello', 6.2], b=1)"



import numpy as np
import pytest

from mcalf.profiles.voigt import (
    double_voigt,
    double_voigt_approx,
    double_voigt_approx_nobg,
    double_voigt_nobg,
    voigt,
    voigt_approx,
    voigt_approx_nobg,
    voigt_nobg,
)

pts1 = np.array([-4.245, -0.324, 0.243, 1.163, 1.739, 99.999])
params1 = [8.242, 0.20, 0.241, 0.129, 5.348, 1.228, 0.152, 0.213, 6.82]

pts2 = np.array([-4.153, -0.323, 0.243, 0.682, 1.739, 99.999])
params2 = [7.142, 0.14, 0.321, 0.274, -4.281, 0.842, 0.181, 0.127, 7.25]


def test_voigt_clib():

    # Test 1
    res = double_voigt(pts1, *params1, clib=True)
    truth = [6.84938173, 9.35261435, 16.43320347, 13.11756603, 8.41536644, 6.82007115]
    assert res == pytest.approx(truth)

    # Test 2
    res = double_voigt(pts2, *params2, clib=True)
    truth = [7.27727702, 10.12250135, 11.49218384, 5.12373520, 7.27801656, 7.25004487]
    assert res == pytest.approx(truth)


def test_voigt_noclib():

    # Test 1
    res = double_voigt(pts1, *params1, clib=False)
    truth = [6.84938217, 9.35261435, 16.43320410, 13.11756590, 8.41536450, 6.82007115]
    assert res == pytest.approx(truth)

    # Test 2
    res = double_voigt(pts2, *params2, clib=False)
    truth = [7.27727751, 10.12250135, 11.49217969, 5.12373519, 7.27801656, 7.25004487]
    assert res == pytest.approx(truth)


def test_voigt_approx():

    # Test 1
    res = double_voigt_approx(pts1, *params1)
    truth = [6.83815735, 8.10237525, 11.61976236, 11.73676887, 8.05422184, 6.82004642]
    assert res == pytest.approx(truth)

    # Test 2
    res = double_voigt_approx(pts2, *params2)
    truth = [7.28207577, 10.29607126, 11.99513939, 7.63595484, 7.41246856, 7.25005616]
    assert res == pytest.approx(truth)


def test_voigt_wrappers():

    for pts, params in ([pts1, params1], [pts2, params2]):

        params_absorption = params[:4]
        params_emission = params[4:-1]
        background = params[-1]

        base_absorption = voigt_nobg(pts, *params_absorption)
        base_emission = voigt_nobg(pts, *params_emission)

        assert all([a == b for a, b in zip(
            voigt(pts, *params_absorption, background),
            base_absorption + background
        )])
        assert all([a == b for a, b in zip(
            double_voigt_nobg(pts, *params_absorption, *params_emission),
            base_absorption + base_emission
        )])
        assert all([a == b for a, b in zip(
            double_voigt(pts, *params_absorption, *params_emission, background),
            base_absorption + base_emission + background
        )])

        base_absorption_approx = voigt_approx_nobg(pts, *params_absorption)
        base_emission_approx = voigt_approx_nobg(pts, *params_emission)

        assert all([a == b for a, b in zip(
            voigt_approx(pts, *params_absorption, background),
            base_absorption_approx + background
        )])
        assert all([a == b for a, b in zip(
            double_voigt_approx_nobg(pts, *params_absorption, *params_emission),
            base_absorption_approx + base_emission_approx
        )])
        assert all([a == b for a, b in zip(
            double_voigt_approx(pts, *params_absorption, *params_emission, background),
            base_absorption_approx + base_emission_approx + background
        )])



import numpy as np
import scipy.interpolate

from mcalf.profiles.gaussian import single_gaussian

__all__ = ['reinterpolate_spectrum', 'normalise_spectrum', 'generate_sigma']


def reinterpolate_spectrum(spectrum, original_wavelengths, constant_wavelengths):
    """Reinterpolate the spectrum.

    Reinterpolates the spectrum such that intensities at `original_wavelengths` are transformed into
    intensities at `constant_wavelengths`. Uses :class:`scipy.interpolate.InterpolatedUnivariateSpline` to
    interpolate.

    Parameters
    ----------
    spectrum : numpy.ndarray, ndim=1
        Spectrum to reinterpolate.
    original_wavelengths : numpy.ndarray, ndim=1, length=length of `spectrum`
        Wavelengths of `spectrum`.
    constant_wavelengths : numpy.ndarray, ndim=1
        Wavelengths to cast `spectrum` into.

    Returns
    -------
    spectrum : numpy.ndarray, length=length of `constant_wavelengths`
        Reinterpolated spectrum.
    """
    s = scipy.interpolate.InterpolatedUnivariateSpline(original_wavelengths, spectrum)  # Fit spline
    return s(constant_wavelengths)  # Evaluate spline at new wavelengths


def normalise_spectrum(spectrum, original_wavelengths=None, constant_wavelengths=None, prefilter_response=None,
                       model=None):
    """Normalise an individual spectrum to have intensities in range [0, 1].

    .. warning::
        Not recommended for normalising many spectra in a loop.

    Parameters
    ----------
    spectrum : numpy.ndarray, ndim=1
        Spectrum to reinterpolate and normalise.
    original_wavelengths : numpy.ndarray, ndim=1, length=length of `spectrum`, optional
        Wavelengths of `spectrum`.
    constant_wavelengths : numpy.ndarray, ndim=1, optional
        Wavelengths to cast `spectrum` into.
    prefilter_response : numpy.ndarray, ndim=1, length=length of `constant_wavelengths`, optional
        Prefilter response to divide spectrum by.
    model : child class of mcalf.models.ModelBase, optional
        Model to extract the above parameters from.

    Returns
    -------
    spectrum : numpy.ndarray, ndim-1, length=length of `constant_wavelengths`
        The normalised spectrum.
    """
    from mcalf.models import ModelBase
    if issubclass(model.__class__, ModelBase):
        if original_wavelengths is None:
            original_wavelengths = model.original_wavelengths
        if constant_wavelengths is None:
            constant_wavelengths = model.constant_wavelengths
        if prefilter_response is None:
            prefilter_response = model.prefilter_response
    if original_wavelengths is not None or constant_wavelengths is not None:
        if original_wavelengths is None or constant_wavelengths is None:
            raise ValueError("original_wavelengths and constant_wavelengths must be provided")
        spectrum = reinterpolate_spectrum(spectrum, original_wavelengths, constant_wavelengths)
    if prefilter_response is not None:
        spectrum /= prefilter_response
    spectrum -= min(spectrum)
    spectrum /= max(spectrum)
    return spectrum


def generate_sigma(sigma_type, wavelengths, line_core, a=-0.95, c=0.04, d=1, centre_rad=7, a_peak=0.4):
    """Generate the default sigma profiles.

    Parameters
    ----------
    sigma_type : int
        Type of profile to generate. Should be either `1` or `2`.
    wavelengths : array_like
        Wavelengths to use for sigma profile.
    line_core : float
        Line core to use as centre of Gaussian sigma profile.
    a : float, optional, default=-0.95
        Amplitude of Gaussian sigma profile.
    c : float, optional, default=0.04
        Sigma of Gaussian sigma profile.
    d : float, optional, default=1
        Background of Gaussian sigma profile.
    centre_rad : int, optional, default=7
        Width of central flattened region.
    a_peak : float, optional, default=0.4
        Amplitude of central 7 pixel section, if `sigma_type` is 2.

    Returns
    -------
    sigma : numpy.ndarray, length=n_wavelengths
        The generated sigma profile.
    """
    sigma = single_gaussian(wavelengths, a, line_core, c, d)  # Define initial sigma profile
    centre_i = np.argmin(sigma)  # Get central index of sigma
    sigma[:centre_i - centre_rad] = sigma[centre_rad:centre_i]  # Shift left half left
    sigma[centre_i + centre_rad:] = sigma[centre_i:-centre_rad]  # Shift right half right
    sigma[centre_i - centre_rad:centre_i + centre_rad] = sigma[centre_i]  # Connect the two halves with a line
    if sigma_type == 2:
        sigma[centre_i - 3:centre_i + 3 + 1] = a_peak  # Lessen importance of peak detail
    return sigma



import os.path

try:
    from setuptools_scm import get_version
    version = get_version(root=os.path.join('..', '..'), relative_to=__file__)
except ImportError:
    from ._version import version  # noqa: F401



import astropy.units
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

__all__ = ['hide_existing_labels', 'calculate_axis_extent', 'calculate_extent', 'class_cmap']


def hide_existing_labels(plot_settings, axes=None, fig=None):
    """Hides labels for each dictionary provided if label already exists in legend.

    Parameters
    ----------
    plot_settings : dict of {str: dict}
        Dictionary of lines to be plotted. Values must be dictionaries with a 'label'
        entry that this function my append with a '_' to hide the label.
    axes : list of matplotlib.axes.Axes, optional, default=None
        List of axes to extract lines labels from. Extracts axes from `fig` if omitted.
    fig : matplotlib.figure.Figure, optional, default=None
        Figure to take line labels from. Uses current figure if omitted.

    Notes
    -----
    Only the ``plot_settings[*]['label']`` values are uses to assess if a label has already
    been used. Other `plot_settings` parameters such as `color` are ignored.

    Examples
    --------

    Import plotting package:

    >>> import matplotlib.pyplot as plt

    Define various plot settings:

    >>> plot_settings = {
    ...     'LineA': {'color': 'r', 'label': 'A'},
    ...     'LineB': {'color': 'g', 'label': 'B'},
    ...     'LineC': {'color': 'b', 'label': 'C'},
    ... }

    Create a figure and plot two lines on the first axes:

    >>> fig, axes = plt.subplots(1, 2)
    >>> axes[0].plot([0, 1], [0, 1], **plot_settings['LineA'])  # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> axes[0].plot([0, 1], [1, 0], **plot_settings['LineB'])  # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]

    Set labels already used to be hidden if used again:

    >>> hide_existing_labels(plot_settings)

    Anything already used will have an underscore prepended:

    >>> [x['label'] for x in plot_settings.values()]
    ['_A', '_B', 'C']

    Plot two lines on the second axes:

    >>> axes[1].plot([0, 1], [0, 1], **plot_settings['LineB'])  # Label hidden  # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> axes[1].plot([0, 1], [1, 0], **plot_settings['LineC'])  # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]

    Show the figure with the legend:

    >>> fig.legend(ncol=3, loc='upper center')  # doctest: +ELLIPSIS
    <matplotlib.legend.Legend object at 0x...>
    >>> plt.show()  # doctest: +SKIP
    >>> plt.close()
    """
    # Get axes:
    if axes is None:
        if fig is None:
            fig = plt.gcf()
        axes = fig.get_axes()

    # Get plotted labels:
    lines = []
    for ax in axes:
        lines.extend(ax.get_lines())
    existing = [line.get_label() for line in lines]

    # Hide labels already plotted:
    for name in plot_settings:
        if plot_settings[name]['label'] in existing:
            plot_settings[name]['label'] = '_' + plot_settings[name]['label']


def calculate_axis_extent(resolution, px, offset=0, unit="Mm"):
    """Calculate the extent from a resolution value along a particular axis.

    Parameters
    ----------
    resolution : float or astropy.units.quantity.Quantity
        Length of each pixel. Unit defaults to `unit` is not an astropy quantity.
    px : int
        Number of pixels extent is being calculated for.
    offset : int or float, default=0
        Number of pixels from the 0 pixel to the first pixel. Defaults to the first
        pixel being at 0 length units. For example, in a 1000 pixel wide dataset,
        setting offset to -500 would place the 0 Mm location at the centre.
    unit : str, default="Mm"
        Default unit string to use if `resolution` is not an astropy quantity.

    Returns
    -------
    first : float
        First extent value.
    last : float
        Last extent value.
    unit : str
        Unit of extent values.
    """

    # Ensure a valid spatial and pixel resolution is provided
    if not isinstance(resolution, (float, astropy.units.quantity.Quantity)):
        raise TypeError('`resolution` values must be either floats or astropy quantities'
                        f', got {type(resolution)}.')
    if not isinstance(px, (int, np.integer)):
        raise TypeError(f'`px` must be an integer, got {type(px)}.')
    if not isinstance(offset, (float, int, np.integer)):
        raise TypeError(f'`offset` must be an float or integer, got {type(offset)}.')

    # Update the default unit if a quantity is provided
    if isinstance(resolution, astropy.units.quantity.Quantity):
        unit = resolution.unit.to_string(astropy.units.format.LatexInline)
        resolution = float(resolution.value)  # Remove the unit

    # Calculate the extent values
    first = offset * resolution
    last = (px + offset) * resolution

    return first, last, unit


def calculate_extent(shape, resolution, offset=(0, 0), ax=None, dimension=None, **kwargs):
    """Calculate the extent from a particular data shape and resolution.

    This function assumes a lower origin is being used with matplotlib.

    Parameters
    ----------
    shape : tuple[int]
        Shape (y, x) of the :class:`numpy.ndarray` of the data being plotted.
        First integer corresponds to the y-axis and the second integer is for the x-axis.
    resolution : tuple[float] or astropy.units.quantity.Quantity
        A 2-tuple (x, y) containing the length of each pixel in the x and y direction respectively.
        If a value has type :class:`astropy.units.quantity.Quantity`, its axis label will
        include its attached unit, otherwise the unit will default to Mm.
        The `ax` parameter must be specified to set its labels.
        If `resolution` is None, this function will immediately return None.
    offset : tuple[float] or int, length=2, optional, default=(0, 0)
        Two offset values (x, y) for the x and y axis respectively.
        Number of pixels from the 0 pixel to the first pixel. Defaults to the first
        pixel being at 0 length units. For example, in a 1000 pixel wide dataset,
        setting offset to -500 would place the 0 Mm location at the centre.
    ax : matplotlib.axes.Axes, optional, default=None
        Axes into which axis labels will be plotted.
        Defaults to not printing axis labels.
    dimension : str or tuple[str] or list[str], length=2, optional, default=None
        If an `ax` (and `resolution`) is provided, use this string as the `dimension name`
        that appears before the ``(unit)`` in the axis label.
        A 2-tuple (x, y) or list [x, y] can instead be given to provide a different name
        for the x-axis and y-axis respectively.
        Defaults is equivalent to ``dimension=('x-axis', 'y-axis')``.
    **kwargs : dict, optional
        Extra keyword arguments to pass to :func:`calculate_axis_extent`.

    Returns
    -------
    extent : tuple[float], length=4
        The extent value that will be passed to matplotlib functions with a lower origin.
        Will return None if `resolution` is None.
    """
    # Calculate a specific extent if a resolution is specified
    if resolution is not None:

        # Validate relevant parameters
        for n, v in (('shape', shape), ('resolution', resolution), ('offset', offset)):
            if not isinstance(v, tuple) or len(v) != 2:
                raise TypeError(f'`{n}` must be a tuple of length 2.')

        # Calculate extent values, and extract units
        ypx, xpx = shape
        l, r, x_unit = calculate_axis_extent(resolution[0], xpx, offset=offset[0], **kwargs)
        b, t, y_unit = calculate_axis_extent(resolution[1], ypx, offset=offset[1], **kwargs)

        # Optionally set the axis labels
        if ax is not None:

            # Extract the dimension name
            if isinstance(dimension, (tuple, list)):  # different value for each dimension
                if len(dimension) != 2:
                    raise TypeError('`dimension` must be a tuple or list of length 2.')
                x_dim = str(dimension[0])
                y_dim = str(dimension[1])
            elif dimension is None:  # default values
                x_dim, y_dim = 'x-axis', 'y-axis'
            elif isinstance(dimension, str):  # single value for both dimensions
                x_dim = y_dim = str(dimension)
            else:
                raise TypeError('`dimension` must be a tuple or list of length 2.')
            ax.set_xlabel(f'{x_dim} ({x_unit})')
            ax.set_ylabel(f'{y_dim} ({y_unit})')

        return l, r, b, t  # extent

    return None  # default extent


def class_cmap(style, n):
    """Create a listed colormap for a specific number of classifications.

    Parameters
    ----------
    style : str
        The named matplotlib colormap to extract a :class:`~matplotlib.colors.ListedColormap`
        from. Colours are selected from `vmin` to `vmax` at equidistant values
        in the range [0, 1]. The :class:`~matplotlib.colors.ListedColormap`
        produced will also show bad classifications and classifications
        out of range in grey.
        The 'original' style is a special case used since early versions
        of this code. It is a hardcoded list of 5 colours. When the number
        of classifications exceeds 5, ``style='viridis'`` will be used.
    n : int
        Number of colours (i.e., number of classifications) to include in
        the colormap.

    Returns
    -------
    cmap : matplotlib.colors.ListedColormap
        Colormap generated for classifications.
    """

    # Validate `n`
    if not isinstance(n, (int, np.integer)):
        raise TypeError(f'`n` must be an integer, got {type(n)}.')

    # Choose colours
    if style == 'original' and n <= 5:  # original colours
        cmap_colors = np.array(['#0072b2', '#56b4e9', '#009e73', '#e69f00', '#d55e00'])[:n]
    else:
        if style == 'original':
            style = 'viridis'  # fallback for >5 classifications
        c = mpl.cm.get_cmap(style)  # query in equal intervals from [0, 1]
        cmap_colors = np.array([c(i / (n - 1)) for i in range(n)])

    # Generate colormap
    cmap = mpl.colors.ListedColormap(cmap_colors)
    cmap.set_over(color='#999999', alpha=1)
    cmap.set_under(color='#999999', alpha=1)

    return cmap
